from ..exceptions import CallosumError


class RPCError(CallosumError):
    """
    A base exception for all RPC-specific errors.
    """

    pass


class RPCUserError(RPCError):
    """
    Represents an error caused in user-defined handlers.
    """

    name: str
    repr: str
    traceback: str

    def __init__(self, name: str, repr_: str, tb: str, *args):
        super().__init__(name, repr_, tb, *args)
        self.name = name
        self.repr = repr_
        self.traceback = tb


class RPCInternalError(RPCError):
    """
    Represents an error caused in Calloum's internal logic.
    """

    name: str
    repr: str
    traceback: str

    def __init__(self, name: str, repr_: str, tb: str, *args):
        super().__init__(name, tb, *args)
        self.name = name
        self.repr = repr_
        self.traceback = tb



from __future__ import annotations

from datetime import datetime
from typing import Any

import attrs
import temporenc
from dateutil.tz import tzutc

from ..abc import (
    AbstractDeserializer,
    AbstractMessage,
    AbstractSerializer,
    RawHeaderBody,
)
from ..serialize import mpackb, munpackb


@attrs.define(frozen=True, slots=True, auto_attribs=True)
class StreamMessage(AbstractMessage):
    # header parts
    created_at: datetime

    # body parts
    body: Any

    @classmethod
    def create(cls, body: Any):
        created_at = datetime.now(tzutc())
        return cls(created_at, body)

    @classmethod
    def decode(
        cls, raw_msg: RawHeaderBody, deserializer: AbstractDeserializer
    ) -> StreamMessage:
        header = munpackb(raw_msg[0])
        created_at = temporenc.unpackb(header[0]).datetime()
        return cls(created_at, deserializer(raw_msg[1]))

    def encode(self, serializer: AbstractSerializer) -> RawHeaderBody:
        header = [
            temporenc.packb(self.created_at),
        ]
        serialized_header: bytes = mpackb(header)
        serialized_body: bytes = serializer(self.body)
        return RawHeaderBody(serialized_header, serialized_body, None)



"""
A simple implementation of RFC 1982 serial number arithmetic
to handle wrapping of packet sequence numbers.

* https://tools.ietf.org/html/rfc1982

Note that when incrementing a serial number, one is allowed
to add up to (2 ** bits - 1) - 1.
"""


def serial_lt(a: int, b: int, bits: int = 32) -> bool:
    half = 2 ** (bits - 1)
    return (a < b and (b - a) < half) or (a > b and (a - b) > half)


def serial_gt(a: int, b: int, bits: int = 32) -> bool:
    half = 2 ** (bits - 1)
    return (a < b and (b - a) > half) or (a > b and (a - b) < half)


def serial_le(a: int, b: int, bits: int = 32) -> bool:
    return a == b or serial_lt(a, b, bits)


def serial_ge(a: int, b: int, bits: int = 32) -> bool:
    return a == b or serial_gt(a, b, bits)



from callosum.serial import serial_gt, serial_lt


def test_serial_comparison():
    assert serial_lt(1, 2, bits=8)
    assert not serial_lt(2, 1, bits=8)

    assert serial_lt(255, 100, bits=8)
    assert serial_lt(255, 126, bits=8)
    assert not serial_lt(255, 127, bits=8)
    assert not serial_lt(255, 200, bits=8)

    assert not serial_gt(1, 2, bits=8)
    assert serial_gt(2, 1, bits=8)

    assert not serial_gt(255, 100, bits=8)
    assert not serial_gt(255, 127, bits=8)
    assert serial_gt(255, 128, bits=8)
    assert serial_gt(255, 200, bits=8)



import asyncio
import logging
import pathlib
import random
import secrets
import sys
import textwrap

import thriftpy2 as thriftpy

from callosum.lower.zeromq import ZeroMQAddress, ZeroMQRPCTransport
from callosum.rpc import Peer, RPCUserError
from callosum.serialize import noop_deserializer, noop_serializer
from callosum.upper.thrift import ThriftClientAdaptor

simple_thrift = thriftpy.load(
    str(pathlib.Path(__file__).parent / "simple.thrift"), module_name="simple_thrift"
)


async def call() -> None:
    peer = Peer(
        connect=ZeroMQAddress("tcp://localhost:5030"),
        transport=ZeroMQRPCTransport,
        serializer=noop_serializer,
        deserializer=noop_deserializer,
        invoke_timeout=2.0,
    )
    adaptor = ThriftClientAdaptor(simple_thrift.SimpleService)
    async with peer:
        response = await peer.invoke("simple", adaptor.echo(secrets.token_hex(16)))
        print(f"echoed {response}")

        response = await peer.invoke("simple", adaptor.echo(secrets.token_hex(16)))
        print(f"echoed {response}")

        a = random.randint(1, 10)
        b = random.randint(10, 20)
        response = await peer.invoke("simple", adaptor.add(a, b))
        print(f"{a} + {b} = {response}")

        try:
            response = await peer.invoke("simple", adaptor.oops())
        except RPCUserError as e:
            print("catched remote error as expected:")
            print(textwrap.indent(e.traceback, prefix="| "))

        try:
            async with asyncio.timeout(0.5):
                await peer.invoke("simple", adaptor.long_delay())
        except asyncio.TimeoutError:
            print(
                "long_delay(): timeout occurred as expected "
                "(with per-call timeout)"
            )
        else:
            print("long_delay(): timeout did not occur!")
            sys.exit(1)

        try:
            await peer.invoke("simple", adaptor.long_delay())
        except asyncio.TimeoutError:
            print(
                "long_delay(): timeout occurred as expected "
                "(with default timeout)"
            )
        else:
            print("long_delay(): timeout did not occur!")
            sys.exit(1)


if __name__ == "__main__":
    logging.basicConfig(
        format="%(asctime)s [%(levelname)s] %(name)s: %(message)s",
        level=logging.INFO,
    )
    log = logging.getLogger()
    asyncio.run(call())



from __future__ import annotations

import asyncio
from typing import Any, AsyncGenerator, Mapping, Optional, Tuple, Union

import attrs
import redis
import redis.asyncio

from ..abc import RawHeaderBody
from ..exceptions import InvalidAddressError
from . import (
    AbstractAddress,
    AbstractBinder,
    AbstractConnection,
    AbstractConnector,
    BaseTransport,
)
from .redis_common import redis_addr_to_url


@attrs.define(auto_attribs=True, slots=True)
class RedisStreamAddress(AbstractAddress):
    """
    Group and consumer are only required by receiver-side.
    Thus they must not be specified for CommonStreamBinder.
    """

    redis_server: Union[str, Tuple[str, int]]
    stream_key: str
    group: Optional[str] = None
    consumer: Optional[str] = None


class DispatchRedisConnection(AbstractConnection):
    __slots__ = ("transport",)

    transport: DispatchRedisTransport
    addr: RedisStreamAddress
    direction_keys: Optional[Tuple[str, str]]

    def __init__(
        self,
        transport: DispatchRedisTransport,
        addr: RedisStreamAddress,
        direction_keys: Optional[Tuple[str, str]] = None,
    ):
        self.transport = transport
        self.addr = addr
        self.direction_keys = direction_keys

    async def recv_message(self) -> AsyncGenerator[Optional[RawHeaderBody], None]:
        assert self.transport._redis is not None
        assert self.addr.group is not None
        assert self.addr.consumer is not None
        if not self.direction_keys:
            stream_key = self.addr.stream_key
        else:
            stream_key = f"{self.addr.stream_key}.{self.direction_keys[0]}"

        try:
            per_key_fetch_list: list[Any] = []
            while not per_key_fetch_list:
                per_key_fetch_list = await self.transport._redis.xreadgroup(
                    self.addr.group,
                    self.addr.consumer,
                    {stream_key: ">"},
                    block=1000,
                )
            for fetch_info in per_key_fetch_list:
                if fetch_info[0].decode() != stream_key:
                    continue
                for item in fetch_info[1]:
                    item_id: bytes = item[0]
                    item_data: dict[bytes, bytes] = item[1]
                    try:
                        if b"meta" in item_data:
                            continue
                        yield RawHeaderBody(
                            item_data[b"hdr"],
                            item_data[b"msg"],
                            None,
                        )
                    finally:
                        await self.transport._redis.xack(
                            stream_key, self.addr.group, item_id
                        )
        except asyncio.CancelledError:
            raise
        except redis.asyncio.ConnectionError:
            yield None

    async def send_message(self, raw_msg: RawHeaderBody) -> None:
        assert self.transport._redis is not None
        if not self.direction_keys:
            stream_key = self.addr.stream_key
        else:
            stream_key = f"{self.addr.stream_key}.{self.direction_keys[1]}"
        await self.transport._redis.xadd(
            stream_key, {b"hdr": raw_msg[0], b"msg": raw_msg[1]}
        )


class DispatchRedisBinder(AbstractBinder):
    """
    DispatchRedisBinder is for use with Publisher class.
    All Publishers using DispatchRedisBinder are supposed
    to provide the same stream key for the purposes
    of subsequent message load-balancing among those,
    who read messages from the stream (Consumers).
    """

    __slots__ = ("transport", "addr", "_addr_url")

    transport: DispatchRedisTransport
    addr: RedisStreamAddress

    async def __aenter__(self):
        self._addr_url = redis_addr_to_url(self.addr.redis_server)
        self.transport._redis = await redis.asyncio.from_url(
            self._addr_url, **self.transport._redis_opts
        )
        key = self.addr.stream_key
        # If there were no stream with the specified key before,
        # it is created as a side effect of adding the message.
        await self.transport._redis.xadd(key, {b"meta": b"create-or-join-to-stream"})
        if self.addr.group:
            raise InvalidAddressError("group")  # group must not be specified
        if self.addr.consumer:
            raise InvalidAddressError("consumer")  # consumer must not be specified
        return DispatchRedisConnection(self.transport, self.addr)

    async def __aexit__(self, exc_type, exc_obj, exc_tb):
        pass


class DispatchRedisConnector(AbstractConnector):
    """
    DispatchRedisConnector is for use with Consumer class.
    All Consumers using DispatchRedisConnector are supposed
    to provide the same stream key and same group name
    for the purposes of subsequent load-balancing. Only the consumer
    names can be different among Consumers.
    Based on the group name, RedisStream will make sure
    that each consumer from the group gets distinct set of messages.
    """

    __slots__ = ("transport", "addr", "_addr_url")

    transport: DispatchRedisTransport
    addr: RedisStreamAddress

    async def __aenter__(self):
        assert self.addr.group is not None
        assert self.addr.consumer is not None
        self._addr_url = redis_addr_to_url(self.addr.redis_server)
        self.transport._redis = await redis.asyncio.from_url(
            self._addr_url, **self.transport._redis_opts
        )
        key = self.addr.stream_key
        # If there were no stream with the specified key before,
        # it is created as a side effect of adding the message.
        await self.transport._redis.xadd(key, {b"meta": b"create-or-join-to-stream"})
        groups = await self.transport._redis.xinfo_groups(key)
        if not any(map(lambda g: g["name"] == self.addr.group.encode(), groups)):
            await self.transport._redis.xgroup_create(key, self.addr.group)
        return DispatchRedisConnection(self.transport, self.addr)

    async def __aexit__(self, exc_type, exc_obj, exc_tb):
        assert self.addr.group is not None
        assert self.addr.consumer is not None
        # we need to create a new Redis connection for cleanup
        # because self.transport._redis gets corrupted upon
        # cancellation of Peer._recv_loop() task.
        _redis = await redis.asyncio.from_url(
            self._addr_url, **self.transport._redis_opts
        )
        try:
            await asyncio.shield(
                _redis.xgroup_delconsumer(
                    self.addr.stream_key, self.addr.group, self.addr.consumer
                )
            )
        finally:
            await _redis.close()


class DispatchRedisTransport(BaseTransport):
    """
    Implementation for unidirectional transport backend by Redis Streams.
    """

    __slots__ = BaseTransport.__slots__ + (
        "_redis_opts",
        "_redis",
    )

    _redis_opts: Mapping[str, Any]
    _redis: Optional[redis.asyncio.Redis]

    binder_cls = DispatchRedisBinder
    connector_cls = DispatchRedisConnector

    def __init__(
        self,
        authenticator,
        **kwargs,
    ) -> None:
        super().__init__(authenticator, **kwargs)
        self._redis_opts = self.transport_opts.get("redis_opts", {})
        self._redis = None

    async def close(self) -> None:
        if self._redis is not None:
            await self._redis.close()



"""
During the testing, you are supposed to launch
multiple consumers. In this way, you can check
whether messages are distributed among them and
whether each consumer gets only the messages
which have not been obtained by others so far.
"""

import asyncio
import json
import logging
import os
import signal

from callosum.lower.dispatch_redis import DispatchRedisTransport, RedisStreamAddress
from callosum.pubsub import Consumer


def handle_heartbeat(msg_body):
    print(f"Heartbeat from agent {msg_body['agent_id']} received.")


async def handle_add(msg_body):
    await asyncio.sleep(2)
    addend1, addend2 = msg_body["addends"]
    sum = addend1 + addend2
    print(f"{addend1} + {addend2} = {sum}")


async def main_handler(msg):
    if msg.body["type"] == "instance_heartbeat":
        handle_heartbeat(msg.body)
    elif msg.body["type"] == "number_addition":
        await handle_add(msg.body)
    else:
        print("InvalidMessageType: message of type EventTypes was expected.")


async def consume() -> None:
    redis_host = os.environ.get("REDIS_HOST", "127.0.0.1")
    redis_port = int(os.environ.get("REDIS_PORT", "6379"))
    cons = Consumer(
        connect=RedisStreamAddress(
            f"redis://{redis_host}:{redis_port}",
            "events",
            "consumer-group",
            "consumer1",
        ),
        deserializer=lambda b: json.loads(b),
        transport=DispatchRedisTransport,
    )
    cons.add_handler(main_handler)

    loop = asyncio.get_running_loop()
    forever = loop.create_future()
    loop.add_signal_handler(signal.SIGINT, forever.cancel)
    loop.add_signal_handler(signal.SIGTERM, forever.cancel)
    async with cons:
        try:
            print("consumer started")
            await forever
        except asyncio.CancelledError:
            pass
    print("consumer terminated")


if __name__ == "__main__":
    logging.basicConfig(
        format="%(asctime)s [%(levelname)s] %(name)s: %(message)s",
        level=logging.INFO,
    )
    log = logging.getLogger()
    asyncio.run(consume())



from __future__ import annotations

import functools

from ..rpc import Peer, RPCMessage


class BaseRPCServerAdaptor:
    __slots__ = ("peer",)

    def __init__(self, peer: Peer):
        self.peer = peer

    async def handle_function(self, request: RPCMessage) -> bytes:
        """
        Implements upper-layer handling of incoming requests and generation of raw
        responses in bytes (= already serialized in the upper layer).
        """
        raise NotImplementedError

    async def handle_stream(self, request: RPCMessage) -> bytes:
        raise NotImplementedError


class BaseRPCClientAdaptor:
    __slots__ = ()

    def __init__(self):
        pass

    async def _call(self, method, args, kwargs):
        """
        Implements upper-layer generation of requests and processing responses.
        """

        # (1) GUIDE: generate raw_request_body (bytes) here
        # (2) Let Callosum to perform send/recv using its lower transport.
        raw_response_body = yield raw_request_body  # noqa
        # (3) GUIDE: parse raw_response_body (bytes) and get the result here
        result = None
        # (4) Return the upper-processed result.
        yield result

    def __getattr__(self, name):
        def _caller(*args, **kwargs):
            return functools.partial(self._call, name, args, kwargs)

        return _caller



import io


class AsyncBytesIO:
    """
    An async wrapper for in-memory bytes stream.
    """

    def __init__(self, initial_data: bytes = b"") -> None:
        self._buf = io.BytesIO(initial_data)

    async def open(self) -> None:
        pass

    def close(self) -> None:
        self._buf.close()

    async def flush(self) -> None:
        pass

    async def readexactly(self, n: int) -> bytes:
        return self._buf.read(n)

    async def read(self, n: int) -> bytes:
        return self._buf.read(n)

    def write(self, val: bytes) -> None:
        self._buf.write(val)

    def getvalue(self) -> bytes:
        return self._buf.getvalue()

    def seek(self, pos: int, flag: int) -> None:
        self._buf.seek(pos, flag)



import yarl


def redis_addr_to_url(
    value: str | tuple[str, int],
    *,
    scheme: str = "redis",
) -> str:
    match value:
        case str():
            url = yarl.URL(value)
            if url.scheme is None:
                return str(yarl.URL(value).with_scheme(scheme))
            return value
        case (host, port):
            return f"{scheme}://{host}:{port}"
        case _:
            raise ValueError("unrecognized address format", value)



from __future__ import annotations

import enum
import sys
import traceback
from typing import Any, Final, Optional

import attrs

try:
    import snappy  # type: ignore

    has_snappy: Final = True
except ImportError:
    has_snappy: Final = False  # type: ignore

from ..abc import (
    AbstractDeserializer,
    AbstractMessage,
    AbstractSerializer,
    RawHeaderBody,
)
from ..exceptions import ConfigurationError
from ..serialize import mpackb, munpackb
from .types import RequestId

# TODO(FUTURE): zero-copy serialization and de-serialization


class TupleEncodingMixin:
    """
    Encodes the class values in order into a msgpack tuple
    and decodes the object from such msgpack tuples.

    The class must be an attrs class.
    """

    @classmethod
    def decode(cls, buffer: bytes) -> Any:
        if not buffer:
            return None
        return cls(*munpackb(buffer))

    def encode(self) -> bytes:
        cls = type(self)
        fields = attrs.fields(cls)  # type: ignore
        values = [getattr(self, f.name) for f in fields]
        return mpackb(values)


class Metadata(TupleEncodingMixin, object):
    """
    Base type for metadata.
    """

    pass


@attrs.define(frozen=True, slots=True)
class FunctionMetadata(Metadata):
    pass


@attrs.define(frozen=True, slots=True)
class ResultMetadata(Metadata):
    pass


@attrs.define(frozen=True, slots=True)
class ErrorMetadata(Metadata):
    name: str
    repr: str
    traceback: str


@attrs.define(frozen=True, slots=True)
class NullMetadata(Metadata):
    pass


class RPCMessageTypes(enum.IntEnum):
    NULL = 0
    FUNCTION = 1  # request for a function
    RESULT = 2  # result of functions
    FAILURE = 3  # error from user handlers
    ERROR = 4  # error from callosum or underlying libraries
    CANCEL = 5  # client-side timeout or cancel request
    CANCELLED = 6  # server-side timeout or cancel request


# mapped from RPCMessageTypes as index
metadata_types = (
    NullMetadata,
    FunctionMetadata,
    ResultMetadata,
    ErrorMetadata,
    ErrorMetadata,  # intended duplication
    NullMetadata,
    NullMetadata,
)


@attrs.define(frozen=True, slots=True, auto_attribs=True)
class RPCMessage(AbstractMessage):
    # transport-layer annotations
    peer_id: Optional[Any]

    # header parts
    msgtype: RPCMessageTypes
    method: str  # function ID
    order_key: str  # replied back as-is
    client_seq_id: int  # replied back as-is

    # body parts (compressable)
    metadata: Optional[Metadata]
    body: Optional[Any]

    @property
    def request_id(self) -> RequestId:
        return (self.method, self.order_key, self.client_seq_id)

    @classmethod
    def result(cls, request, result_body):
        """
        Creates an RPCMessage instance represents a execution result.
        """
        return cls(
            request.peer_id,
            RPCMessageTypes.RESULT,
            request.method,
            request.order_key,
            request.client_seq_id,
            ResultMetadata(),
            result_body,
        )

    @classmethod
    def failure(cls, request):
        """
        Creates an RPCMessage instance containing exception information,
        when the exception is from user-defined handlers or upper adaptation layers.

        It must be called in an exception handler context, where ``sys.exc_info()``
        returns a non-null tuple.
        """
        exc_info = sys.exc_info()
        return cls(
            request.peer_id,
            RPCMessageTypes.FAILURE,
            request.method,
            request.order_key,
            request.client_seq_id,
            ErrorMetadata(
                exc_info[0].__name__,
                repr(exc_info[1]),
                traceback.format_exc(),
            ),
            None,
        )

    @classmethod
    def error(cls, request):
        """
        Creates an RPCMessage instance containing exception information,
        when the exception is from Callosum's internals.

        It must be called in an exception handler context, where ``sys.exc_info()``
        returns a non-null tuple.
        """
        exc_info = sys.exc_info()
        return cls(
            request.peer_id,
            RPCMessageTypes.ERROR,
            request.method,
            request.order_key,
            request.client_seq_id,
            ErrorMetadata(
                exc_info[0].__name__,
                repr(exc_info[1]),
                traceback.format_exc(),
            ),
            None,
        )

    @classmethod
    def cancel(cls, request):
        """
        Creates an RPCMessage instance represents a cancellation of
        the given request triggered from the client-side.
        """
        return cls(
            request.peer_id,
            RPCMessageTypes.CANCEL,
            request.method,
            request.order_key,
            request.client_seq_id,
            NullMetadata(),
            None,
        )

    @classmethod
    def cancelled(cls, request):
        """
        Creates an RPCMessage instance represents a cancellation of
        the given request triggered from the server-side.
        """
        return cls(
            request.peer_id,
            RPCMessageTypes.CANCELLED,
            request.method,
            request.order_key,
            request.client_seq_id,
            NullMetadata(),
            None,
        )

    @classmethod
    def decode(
        cls, raw_msg: RawHeaderBody, deserializer: AbstractDeserializer
    ) -> RPCMessage:
        header = munpackb(raw_msg.header)
        msgtype = RPCMessageTypes(header["type"])
        compressed = header["zip"]
        raw_data = raw_msg.body
        if compressed:
            if not has_snappy:
                raise ConfigurationError("python-snappy is not installed")
            raw_data = snappy.decompress(raw_data)
        data = munpackb(raw_data)
        metadata = metadata_types[msgtype].decode(data["meta"])
        if msgtype in (RPCMessageTypes.FUNCTION, RPCMessageTypes.RESULT):
            body = deserializer(data["body"])
        else:
            body = data["body"]
        return cls(
            raw_msg.peer_id,
            msgtype,
            header["meth"],
            header["okey"],
            header["seq"],
            metadata,
            body,
        )

    def encode(
        self, serializer: AbstractSerializer, compress: bool = False
    ) -> RawHeaderBody:
        metadata = b""
        if self.metadata is not None:
            metadata = self.metadata.encode()
        header = {
            "type": int(self.msgtype),
            "meth": self.method,
            "okey": self.order_key,
            "seq": self.client_seq_id,
            "zip": compress,
        }
        serialized_header: bytes = mpackb(header)
        body: Optional[bytes]
        if self.msgtype in (RPCMessageTypes.FUNCTION, RPCMessageTypes.RESULT):
            body = serializer(self.body)
        else:
            body = self.body
        data = {
            "meta": metadata,
            "body": body,
        }
        serialized_data: bytes = mpackb(data)
        if compress:
            if not has_snappy:
                raise ConfigurationError("python-snappy is not installed")
            serialized_data = snappy.compress(serialized_data)
        return RawHeaderBody(
            serialized_header,
            serialized_data,
            self.peer_id,
        )



import os
import re
from pathlib import Path


def main():
    gh_env_file = Path(os.environ.get("GITHUB_ENV", "/dev/null"))
    version = Path("version.txt").read_text().strip()
    m = re.search(r"(rc\d+|a\d+|b\d+|dev\d+)$", version)
    is_prerelease = "true" if m is not None else "false"
    with open(gh_env_file, "a") as f:
        f.write(f"IS_PRERELEASE={is_prerelease}\n")


if __name__ == "__main__":
    main()



from __future__ import annotations

import asyncio
import functools
import logging
from contextlib import aclosing
from typing import Any, List, Mapping, Optional, Type, Union

from ..abc import (
    AbstractChannel,
    AbstractDeserializer,
    AbstractSerializer,
    QueueSentinel,
)
from ..auth import AbstractClientAuthenticator, AbstractServerAuthenticator
from ..lower import (
    AbstractAddress,
    AbstractBinder,
    AbstractConnection,
    AbstractConnector,
    BaseTransport,
)
from .message import StreamMessage
from .types import ConsumerCallback

log = logging.getLogger(__spec__.name)  # type: ignore[name-defined]


class Publisher(AbstractChannel):
    """
    Represents a unidirectional message publisher.
    """

    _connection: Optional[AbstractConnection]
    _opener: Optional[AbstractBinder]
    _outgoing_queue: asyncio.Queue[Union[QueueSentinel, StreamMessage]]
    _send_task: Optional[asyncio.Task]
    _serializer: AbstractSerializer

    def __init__(
        self,
        *,
        serializer: AbstractSerializer,
        bind: Optional[AbstractAddress] = None,
        transport: Optional[Type[BaseTransport]] = None,
        authenticator: Optional[
            AbstractClientAuthenticator | AbstractServerAuthenticator
        ] = None,
        transport_opts: Mapping[str, Any] = {},
    ) -> None:
        if bind is None:
            raise ValueError("You must specify the bind address.")
        self._bind = bind
        self._opener = None
        self._connection = None
        self._serializer = serializer
        if transport is None:
            raise ValueError("You must provide a transport class.")
        self._transport = transport(
            authenticator=authenticator, transport_opts=transport_opts
        )

        self._outgoing_queue = asyncio.Queue()
        self._send_task = None

        self._log = logging.getLogger(__name__ + ".Publisher")

    async def _send_loop(self) -> None:
        if self._connection is None:
            raise RuntimeError("consumer is not opened yet.")
        while True:
            try:
                msg = await self._outgoing_queue.get()
                if msg is QueueSentinel.CLOSED:
                    break
                await self._connection.send_message(msg.encode(self._serializer))
            except asyncio.CancelledError:
                break
            except Exception:
                log.exception("unexpected error")

    async def __aenter__(self) -> Publisher:
        _opener = functools.partial(self._transport.bind, self._bind)()
        self._opener = _opener
        self._connection = await _opener.__aenter__()
        self._send_task = asyncio.create_task(self._send_loop())
        return self

    async def __aexit__(self, *exc_info) -> None:
        if self._send_task is not None:
            if self._opener is not None:
                await self._opener.__aexit__(None, None, None)
            await self._outgoing_queue.put(QueueSentinel.CLOSED)
            await self._send_task
        if self._transport is not None:
            await self._transport.close()

    def push(self, body) -> None:
        msg = StreamMessage.create(body)
        self._outgoing_queue.put_nowait(msg)


class Consumer(AbstractChannel):
    """
    Represents a unidirectional message consumer.
    """

    _connection: Optional[AbstractConnection]
    _opener: Optional[AbstractConnector]
    _incoming_queue: asyncio.Queue[StreamMessage]
    _recv_task: Optional[asyncio.Task]
    _deserializer: AbstractDeserializer
    _handler_registry: List[ConsumerCallback]

    def __init__(
        self,
        *,
        deserializer: AbstractDeserializer,
        connect: Optional[AbstractAddress] = None,
        transport: Optional[Type[BaseTransport]] = None,
        authenticator: Optional[
            AbstractClientAuthenticator | AbstractServerAuthenticator
        ] = None,
        transport_opts: Mapping[str, Any] = {},
        scheduler=None,
        max_concurrency: int = 100,
    ) -> None:
        if connect is None:
            raise ValueError("You must specify the connect address.")
        self._connect = connect
        self._opener = None
        self._connection = None
        self._deserializer = deserializer
        if transport is None:
            raise ValueError("You must provide a transport class.")
        self._transport = transport(
            authenticator=authenticator, transport_opts=transport_opts
        )
        self._scheduler = scheduler
        self._concurrency_sema = asyncio.Semaphore(max_concurrency)

        self._incoming_queue = asyncio.Queue()
        self._handler_registry = []
        self._recv_task = None

        self._log = logging.getLogger(__name__ + ".Consumer")

    def add_handler(self, callback: ConsumerCallback) -> None:
        self._handler_registry.append(callback)

    async def _task_wrapper(self, handler: ConsumerCallback, msg: Any) -> None:
        async with self._concurrency_sema:
            await handler(msg)

    async def _recv_loop(self) -> None:
        if self._connection is None:
            raise RuntimeError("consumer is not opened yet.")
        try:
            async with aclosing(self._connection.recv_message()) as agen:
                async for raw_msg in agen:
                    if raw_msg is None:
                        return
                    msg = StreamMessage.decode(raw_msg, self._deserializer)
                    for handler in self._handler_registry:
                        if asyncio.iscoroutinefunction(handler):
                            asyncio.create_task(self._task_wrapper(handler, msg))
                            # TODO: keep weak-refs to tasks and use it to
                            #       cancel running tasks upon shutdown
                        else:
                            handler(msg)
        except asyncio.CancelledError:
            pass
        except Exception:
            log.exception("unexpected error")

    async def __aenter__(self) -> Consumer:
        _opener = functools.partial(self._transport.connect, self._connect)()
        self._opener = _opener
        self._connection = await _opener.__aenter__()
        self._recv_task = asyncio.create_task(self._recv_loop())
        return self

    async def __aexit__(self, *exc_info) -> None:
        if self._recv_task is not None:
            if self._opener is not None:
                await self._opener.__aexit__(None, None, None)
            self._recv_task.cancel()
            await self._recv_task
        if self._transport is not None:
            await self._transport.close()



import asyncio
import json
import logging
import os
import signal
import time

from callosum.lower.rpc_redis import RedisStreamAddress, RPCRedisTransport
from callosum.rpc import Peer


async def handle_echo(request):
    # NOTE: Adding this part with "await asyncio.sleep(1)" breaks the code.
    time.sleep(1.0)
    print("After sleeping")
    return {
        "received": request.body["sent"],
    }


async def handle_add(request):
    return {
        "result": request.body["a"] + request.body["b"],
    }


async def serve() -> None:
    redis_host = os.environ.get("REDIS_HOST", "127.0.0.1")
    redis_port = int(os.environ.get("REDIS_PORT", "6379"))
    peer = Peer(
        bind=RedisStreamAddress(
            f"redis://{redis_host}:{redis_port}",
            "myservice",
            "server-group",
            "client1",
        ),
        transport=RPCRedisTransport,
        serializer=lambda o: json.dumps(o).encode("utf8"),
        deserializer=lambda b: json.loads(b),
    )
    peer.handle_function("echo", handle_echo)
    peer.handle_function("add", handle_add)

    loop = asyncio.get_running_loop()
    forever = loop.create_future()
    loop.add_signal_handler(signal.SIGINT, forever.cancel)
    loop.add_signal_handler(signal.SIGTERM, forever.cancel)
    async with peer:
        try:
            print("server started")
            await forever
        except asyncio.CancelledError:
            pass
    print("server terminated")


if __name__ == "__main__":
    logging.basicConfig(
        format="%(asctime)s [%(levelname)s] %(name)s: %(message)s",
        level=logging.INFO,
    )
    log = logging.getLogger()
    asyncio.run(serve())



from .channel import Consumer, Publisher
from .message import StreamMessage
from .types import ConsumerCallback

__all__ = (
    "StreamMessage",
    "Publisher",
    "Consumer",
    "ConsumerCallback",
)



import asyncio
import logging
import pathlib
import signal

import thriftpy2 as thriftpy

from callosum.lower.zeromq import ZeroMQAddress, ZeroMQRPCTransport
from callosum.rpc import Peer
from callosum.serialize import noop_deserializer, noop_serializer
from callosum.upper.thrift import ThriftServerAdaptor

simple_thrift = thriftpy.load(
    str(pathlib.Path(__file__).parent / "simple.thrift"), module_name="simple_thrift"
)


class SimpleDispatcher:
    async def echo(self, msg: str) -> str:
        return msg

    async def add(self, a: int, b: int) -> int:
        return a + b

    async def oops(self) -> bool:
        raise ZeroDivisionError("oops")

    async def long_delay(self) -> bool:
        await asyncio.sleep(5.0)
        return True


async def serve() -> None:
    peer = Peer(
        bind=ZeroMQAddress("tcp://127.0.0.1:5030"),
        serializer=noop_serializer,
        deserializer=noop_deserializer,
        transport=ZeroMQRPCTransport,
    )
    adaptor = ThriftServerAdaptor(
        peer, simple_thrift.SimpleService, SimpleDispatcher()
    )
    peer.handle_function("simple", adaptor.handle_function)

    loop = asyncio.get_running_loop()
    forever = loop.create_future()
    loop.add_signal_handler(signal.SIGINT, forever.cancel)
    loop.add_signal_handler(signal.SIGTERM, forever.cancel)
    async with peer:
        try:
            print("server started")
            await forever
        except asyncio.CancelledError:
            pass
    print("server terminated")


if __name__ == "__main__":
    logging.basicConfig(
        format="%(asctime)s [%(levelname)s] %(name)s: %(message)s",
        level=logging.INFO,
    )
    log = logging.getLogger()
    asyncio.run(serve())



from __future__ import annotations

import abc
from typing import Any, AsyncGenerator, ClassVar, Mapping, Optional, Type

from ..abc import RawHeaderBody
from ..auth import AbstractClientAuthenticator, AbstractServerAuthenticator


class AbstractMessagingMixin(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def recv_message(self) -> AsyncGenerator[Optional[RawHeaderBody], None]:
        yield None

    @abc.abstractmethod
    async def send_message(self, raw_msg: RawHeaderBody) -> None:
        raise NotImplementedError


class AbstractAddress:
    pass


class AbstractStreamingMixin(metaclass=abc.ABCMeta):
    # TODO: define
    pass


class AbstractConnection(
    AbstractMessagingMixin, AbstractStreamingMixin, metaclass=abc.ABCMeta
):
    """
    An abstract interface for communication operations, except its lifecycle
    management operations which are responsible to binder and connector.
    """

    pass


class AbstractBinder(metaclass=abc.ABCMeta):
    __slots__ = ("transport", "addr")

    def __init__(
        self,
        transport: BaseTransport,
        addr: AbstractAddress,
        **transport_opts,
    ) -> None:
        self.transport = transport
        self.addr = addr

    @abc.abstractmethod
    async def __aenter__(self) -> AbstractConnection:
        """
        Create a listening connection bound on self.addr.
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def __aexit__(self, exc_type, exc_obj, exc_tb):
        """
        Close the listening connection.
        """
        pass


class AbstractConnector(metaclass=abc.ABCMeta):
    __slots__ = ("transport", "addr")

    def __init__(
        self,
        transport: BaseTransport,
        addr: AbstractAddress,
        **transport_opts,
    ) -> None:
        self.transport = transport
        self.addr = addr

    @abc.abstractmethod
    async def __aenter__(self) -> AbstractConnection:
        """
        Return a connection to self.addr.
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def __aexit__(self, exc_type, exc_obj, exc_tb):
        """
        Close/release the connection.
        """
        pass


class BaseTransport(metaclass=abc.ABCMeta):
    binder_cls: ClassVar[Type[AbstractBinder]]
    connector_cls: ClassVar[Type[AbstractConnector]]

    __slots__ = (
        "authenticator",
        "transport_opts",
    )
    authenticator: Optional[
        AbstractClientAuthenticator | AbstractServerAuthenticator
    ]
    transport_opts: Mapping[str, Any]

    def __init__(
        self,
        authenticator: Optional[
            AbstractClientAuthenticator | AbstractServerAuthenticator
        ] = None,
        *,
        transport_opts: Optional[Mapping[str, Any]] = None,
        **kwargs,
    ) -> None:
        self.authenticator = authenticator
        self.transport_opts = transport_opts or {}

    def bind(self, bind_addr: AbstractAddress) -> AbstractBinder:
        return type(self).binder_cls(self, bind_addr, **self.transport_opts)

    def connect(self, connect_addr: AbstractAddress) -> AbstractConnector:
        return type(self).connector_cls(self, connect_addr, **self.transport_opts)

    async def close(self) -> None:
        """
        Close all open connections and release system resources.
        This may be left empty for transports without connection pooling or
        persistent connections.
        """
        pass



import asyncio
import json
import logging

from callosum.lower.zeromq import ZeroMQAddress, ZeroMQRPCTransport
from callosum.rpc import Peer


async def call() -> None:
    peer = Peer(
        connect=ZeroMQAddress("tcp://localhost:5010"),
        transport=ZeroMQRPCTransport,
        serializer=lambda o: json.dumps(o).encode("utf8"),
        deserializer=lambda b: json.loads(b),
        invoke_timeout=5.0,
    )
    async with peer:
        print("Check the server log to see in which order echo/add are executed.\n")

        print("== Calling with the same ordering key ==")
        tasks = [
            peer.invoke("echo", {"sent": "bbbb"}, order_key="mykey"),
            peer.invoke("add", {"a": 1, "b": 2}, order_key="mykey"),
        ]
        responses = await asyncio.gather(*tasks)
        print(responses)

        await peer.invoke("print_delim", {})

        print("== Calling without any ordering key ==")
        tasks = [
            peer.invoke("echo", {"sent": "bbbb"}),
            peer.invoke("add", {"a": 1, "b": 2}),
        ]
        responses = await asyncio.gather(*tasks)
        print(responses)

        await peer.invoke("print_delim", {})


if __name__ == "__main__":
    logging.basicConfig(
        format="%(asctime)s [%(levelname)s] %(name)s: %(message)s",
        level=logging.INFO,
    )
    log = logging.getLogger()
    asyncio.run(call())



import asyncio
import io
import logging
from typing import Any, Optional, Sequence

from thriftpy2.contrib.aio.processor import TAsyncProcessor
from thriftpy2.contrib.aio.protocol.binary import TAsyncBinaryProtocol
from thriftpy2.thrift import TApplicationException, TMessageType, args_to_kwargs

from ..io import AsyncBytesIO
from ..rpc import Peer, RPCMessage
from . import BaseRPCClientAdaptor, BaseRPCServerAdaptor

logger = logging.getLogger(__spec__.name)  # type: ignore[name-defined]


class ThriftServerAdaptor(BaseRPCServerAdaptor):
    __slots__ = BaseRPCServerAdaptor.__slots__ + (
        "_processor",
        "_protocol_cls",
        "_exec_timeout",
    )

    def __init__(
        self, peer: Peer, service, handler, *, exec_timeout: Optional[float] = None
    ) -> None:
        super().__init__(peer)
        self._processor = TAsyncProcessor(service, handler)
        self._protocol_cls = TAsyncBinaryProtocol
        self._exec_timeout = exec_timeout

    async def handle_function(self, request: RPCMessage) -> bytes:
        assert request.body is not None
        reader_trans = AsyncBytesIO(request.body)
        writer_trans = AsyncBytesIO()
        iproto = self._protocol_cls(reader_trans)
        oproto = self._protocol_cls(writer_trans)
        try:
            async with asyncio.timeout(self._exec_timeout):
                await self._processor.process(iproto, oproto)
        except (asyncio.IncompleteReadError, ConnectionError):
            logger.debug("client has closed the connection")
            writer_trans.close()
        except asyncio.TimeoutError:
            logger.debug("timeout when processing the client request")
            writer_trans.close()
            raise
        except asyncio.CancelledError:
            logger.debug("service cancelled")
            writer_trans.close()
            raise
        except Exception:
            writer_trans.close()
            raise
        finally:
            reader_trans.close()
        # Even with oneway requests,
        # Callosum will respond with an empty body.
        response_body = writer_trans.getvalue()
        writer_trans.close()
        return response_body


class ThriftClientAdaptor(BaseRPCClientAdaptor):
    __slots__ = BaseRPCClientAdaptor.__slots__ + (
        "_service",
        "_protocol_cls",
        "_seqid",
    )

    def __init__(self, service):
        super().__init__()
        self._service = service
        self._protocol_cls = TAsyncBinaryProtocol
        self._seqid = 0

    async def _call(self, method: str, args: Sequence[Any], kwargs):
        reader_trans = AsyncBytesIO()
        writer_trans = AsyncBytesIO()
        iproto = self._protocol_cls(reader_trans)
        oproto = self._protocol_cls(writer_trans)
        try:
            self._seqid += 1

            # Write to the output buffer with Thrift.
            kw = args_to_kwargs(
                getattr(self._service, method + "_args").thrift_spec, *args
            )
            kwargs.update(kw)
            oproto.write_message_begin(method, TMessageType.CALL, self._seqid)
            api_args = getattr(self._service, method + "_args")()
            for k, v in kwargs.items():
                setattr(api_args, k, v)
            api_args.write(oproto)
            oproto.write_message_end()

            # Switch over the control to Callosum so that it can perform send/recv
            # with its own lower transport layer.
            raw_request_body = writer_trans.getvalue()
            raw_response_body = yield raw_request_body
            if raw_response_body is None:
                yield None
                return
            reader_trans.write(raw_response_body)
            reader_trans.seek(0, io.SEEK_SET)

            # Now read the response buffer with Thrift.
            fname, mtype, rseqid = await iproto.read_message_begin()
            if rseqid != self._seqid:
                raise TApplicationException(
                    TApplicationException.BAD_SEQUENCE_ID,
                    fname + " failed: out of sequence response",
                )

            if mtype == TMessageType.EXCEPTION:
                x = TApplicationException()
                await iproto.read_struct(x)
                await iproto.read_message_end()
                raise x
            result = getattr(self._service, method + "_result")()
            await iproto.read_struct(result)
            await iproto.read_message_end()

            if hasattr(result, "success") and result.success is not None:
                yield result.success
            if len(result.thrift_spec) == 0:
                yield None
            for k, v in result.__dict__.items():
                if k != "success" and v:
                    raise v
        finally:
            writer_trans.close()
            reader_trans.close()



from typing import Any, Optional

import msgpack


def mpackb(v: Any) -> bytes:
    return msgpack.packb(v, use_bin_type=True)


def munpackb(b: bytes) -> Any:
    return msgpack.unpackb(b, raw=False, use_list=False)


def noop_serializer(o: Optional[Any]) -> bytes:
    assert isinstance(o, bytes)
    return o


def noop_deserializer(b: bytes) -> Optional[Any]:
    return b



import json
from contextlib import asynccontextmanager
from typing import AsyncIterator, Optional, Tuple

import pytest
import zmq

from callosum.auth import (
    AbstractClientAuthenticator,
    AbstractServerAuthenticator,
    AuthResult,
    Credential,
    Identity,
    create_keypair,
)
from callosum.exceptions import AuthenticationError
from callosum.lower.zeromq import ZeroMQAddress, ZeroMQRPCTransport
from callosum.ordering import ExitOrderedAsyncScheduler
from callosum.rpc import Peer


async def handle_echo(request):
    return {
        "received": request.body["sent"],
    }


server_keypair = create_keypair()
client1_keypair = create_keypair()
client2_keypair = create_keypair()


class MyServerAuthenticator(AbstractServerAuthenticator):
    def __init__(self, domain: str, server_keypair: Tuple[bytes, bytes]) -> None:
        self.domain = domain
        self._public_key = server_keypair[0]
        self._private_key = server_keypair[1]
        self.user_db = {
            client1_keypair[0]: "myuser1",
        }

    async def server_identity(self) -> Identity:
        return Identity(domain=self.domain, private_key=self._private_key)

    async def check_client(self, creds: Credential) -> AuthResult:
        if creds.domain != self.domain:
            return AuthResult(success=False)
        if user_id := self.user_db.get(creds.public_key, None):
            return AuthResult(
                success=True,
                user_id=user_id,
            )
        return AuthResult(success=False)

    async def server_public_key(self) -> bytes:
        return self._public_key

    async def client_identity(self) -> Identity:
        raise NotImplementedError

    async def client_public_key(self) -> bytes:
        raise NotImplementedError


class MyClientAuthenticator(AbstractClientAuthenticator):
    def __init__(
        self,
        domain: str,
        keypair: Tuple[bytes, bytes],
        server_public_key: bytes,
    ) -> None:
        self.domain = domain
        self._public_key = keypair[0]
        self._private_key = keypair[1]
        self._server_public_key = server_public_key

    async def server_identity(self) -> Identity:
        raise NotImplementedError

    async def check_client(self, creds: Credential) -> AuthResult:
        raise NotImplementedError

    async def server_public_key(self) -> bytes:
        return self._server_public_key

    async def client_identity(self) -> Identity:
        return Identity(domain=self.domain, private_key=self._private_key)

    async def client_public_key(self) -> bytes:
        return self._public_key


@asynccontextmanager
async def server(
    auth: Optional[AbstractServerAuthenticator],
) -> AsyncIterator[str]:
    scheduler = ExitOrderedAsyncScheduler()
    # Bind to a random port to avoid conflicts between different test cases.
    peer = Peer(
        bind=ZeroMQAddress("tcp://127.0.0.1:*"),
        transport=ZeroMQRPCTransport,
        authenticator=auth,
        scheduler=scheduler,
        serializer=lambda o: json.dumps(o).encode("utf8"),
        deserializer=lambda b: json.loads(b),
    )
    peer.handle_function("echo", handle_echo)
    async with peer:
        # TODO: make this a public API of callosum
        underlying_sock = peer._transport._sock  # type: ignore
        endpoint = underlying_sock.getsockopt(zmq.LAST_ENDPOINT)
        print("server started")
        yield endpoint
    print("server terminated")


async def client(
    endpoint: str,
    auth: Optional[AbstractClientAuthenticator],
) -> None:
    peer = Peer(
        connect=ZeroMQAddress(endpoint),
        transport=ZeroMQRPCTransport,
        authenticator=auth,
        serializer=lambda o: json.dumps(o).encode("utf8"),
        deserializer=lambda b: json.loads(b),
        invoke_timeout=2.0,
        transport_opts={"handshake_timeout": 0.2},
    )
    async with peer:
        resp = await peer.invoke("echo", {"sent": "asdf"})
        assert resp["received"] == "asdf"


@pytest.mark.asyncio
async def test_auth_null():
    async with server(
        None,
    ) as endpoint:
        await client(endpoint, None)


@pytest.mark.asyncio
async def test_auth_good_user():
    async with server(
        MyServerAuthenticator("testing", server_keypair),
    ) as endpoint:
        await client(
            endpoint,
            MyClientAuthenticator(
                "testing",
                client1_keypair,
                server_keypair[0],
            ),
        )


@pytest.mark.asyncio
async def test_auth_wrong_user():
    async with server(
        MyServerAuthenticator("testing", server_keypair),
    ) as endpoint:
        with pytest.raises(AuthenticationError):
            await client(
                endpoint,
                MyClientAuthenticator(
                    "testing",
                    client2_keypair,
                    server_keypair[0],
                ),
            )


@pytest.mark.asyncio
async def test_auth_no_auth():
    async with server(
        MyServerAuthenticator("testing", server_keypair),
    ) as endpoint:
        with pytest.raises(AuthenticationError):
            await client(endpoint, None)



from __future__ import annotations

import asyncio
import contextlib
import logging
import secrets
import time
from typing import Any, AsyncGenerator, ClassVar, Mapping, Optional, Type

import attrs
import zmq
import zmq.asyncio
import zmq.constants
import zmq.utils.monitor
from zmq.utils import z85

from callosum.exceptions import AuthenticationError

from ..abc import RawHeaderBody
from ..auth import (
    AbstractClientAuthenticator,
    AbstractServerAuthenticator,
    Credential,
)
from . import (
    AbstractAddress,
    AbstractBinder,
    AbstractConnection,
    AbstractConnector,
    BaseTransport,
)

ZAP_VERSION = b"1.0"

_default_zsock_opts = {
    zmq.LINGER: 100,
}
log = logging.getLogger(__spec__.name)  # type: ignore[name-defined]


@attrs.define(auto_attribs=True, slots=True)
class ZeroMQAddress(AbstractAddress):
    uri: str


class ZAPServer:
    """
    A simple authenticator adaptor that implements ZAP protocol.
    """

    _log: logging.Logger
    _zap_socket: Optional[zmq.asyncio.Socket]
    _zctx: zmq.asyncio.Context

    def __init__(
        self,
        zctx: zmq.asyncio.Context,
        authenticator: Optional[AbstractServerAuthenticator] = None,
    ) -> None:
        self._log = logging.getLogger(__name__ + ".ZAPServer")
        self._authenticator = authenticator
        self._zap_socket = None
        self._zctx = zctx

    async def serve(self) -> None:
        self._zap_socket = self._zctx.socket(zmq.REP)
        self._zap_socket.linger = 1
        self._zap_socket.bind("inproc://zeromq.zap.01")
        try:
            while True:
                try:
                    msg = await self._zap_socket.recv_multipart()
                    await self.handle_zap_message(msg)
                except Exception:
                    self._log.exception("unexpected error in the ZAPServer handler")
        except asyncio.CancelledError:
            pass
        finally:
            self._zap_socket.close()
            self._zap_socket = None

    def close(self) -> None:
        if self._zap_socket is not None:
            self._zap_socket.close()
        self._zap_socket = None

    async def handle_zap_message(self, msg) -> None:
        if len(msg) < 6:
            self._log.error("Invalid ZAP message, not enough frames: %r", msg)
            if len(msg) < 2:
                self._log.error("Not enough information to reply")
            else:
                await self._send_zap_reply(msg[1], b"400", b"Not enough frames")
            return

        version, request_id, domain, address, sock_identity, mechanism = msg[:6]
        self._log.debug("zap message %s", msg)
        credentials = msg[6:]
        domain = domain.decode("utf8", "replace")
        address = address.decode("utf8", "replace")

        if version != ZAP_VERSION:
            self._log.error("Invalid ZAP version: %r", msg)
            await self._send_zap_reply(request_id, b"400", b"Invalid version")
            return

        self._log.debug(
            "version: %r, request_id: %r, domain: %r, "
            "address: %r, sock_identity: %r, mechanism: %r",
            version,
            request_id,
            domain,
            address,
            sock_identity,
            mechanism,
        )
        allowed = False
        user_id = "<anonymous>"
        reason = b"Access denied"

        if self._authenticator is None:
            # When authenticator is not set, allow any NULL auth-requests.
            await self._send_zap_reply(request_id, b"200", b"OK", user_id)
            return

        if mechanism == b"CURVE":
            # For CURVE, even a whitelisted address must authenticate
            if len(credentials) != 1:
                self._log.warning("Invalid CURVE credentials: %r", credentials)
                await self._send_zap_reply(
                    request_id, b"400", b"Invalid credentials"
                )
                return
            client_pubkey = z85.encode(credentials[0])
            creds = Credential(domain, client_pubkey)
            result = await self._authenticator.check_client(creds)
            self._log.info(
                "authentication result with credentials %r -> %r",
                creds,
                result,
            )
            allowed = result.success
            if allowed:
                assert (
                    result.user_id is not None
                ), "expected valid user ID from check_client() callback"
                user_id = result.user_id
        else:
            # In Callosum, we only support public-key based authentication.
            # Other possible values: b'NULL', b'PLAIN', b'GSSAPI'
            reason = b"Unsupported authentication mechanism"

        if allowed:
            assert user_id is not None
            await self._send_zap_reply(request_id, b"200", b"OK", user_id)
        else:
            await self._send_zap_reply(request_id, b"400", reason)

    async def _send_zap_reply(
        self,
        request_id: bytes,
        status_code: bytes,
        status_text: bytes,
        user_id: str = "",
    ) -> None:
        metadata = b""
        self._log.debug("ZAP reply code=%s text=%s", status_code, status_text)
        reply = (
            ZAP_VERSION,
            request_id,
            status_code,
            status_text,
            user_id.encode("utf8", "replace"),
            metadata,
        )
        assert self._zap_socket is not None
        await self._zap_socket.send_multipart(reply)


async def init_authenticator(
    authenticator: AbstractServerAuthenticator | AbstractClientAuthenticator | None,
    sock: zmq.asyncio.Socket,
) -> None:
    match authenticator:
        case AbstractServerAuthenticator() as auth:
            server_public_key = await auth.server_public_key()
            server_id = await auth.server_identity()
            sock.zap_domain = server_id.domain.encode("utf8")
            sock.setsockopt(zmq.CURVE_SERVER, 1)
            sock.setsockopt(zmq.CURVE_SECRETKEY, server_id.private_key)
            log.debug(
                "init_authenticator(): start as server (server_public_key=%r)",
                server_public_key.decode(),
            )
        case AbstractClientAuthenticator() as auth:
            client_id = await auth.client_identity()
            client_public_key = await auth.client_public_key()
            server_public_key = await auth.server_public_key()
            sock.zap_domain = client_id.domain.encode("utf8")
            sock.setsockopt(zmq.CURVE_SERVERKEY, server_public_key)
            sock.setsockopt(zmq.CURVE_PUBLICKEY, client_public_key)
            sock.setsockopt(zmq.CURVE_SECRETKEY, client_id.private_key)
            log.debug(
                (
                    "init_authenticator(): start as client "
                    "(client_public_key=%r, server_public_key=%r)"
                ),
                client_public_key.decode(),
                server_public_key.decode(),
            )
        case None:
            pass


class ZeroMQRPCConnection(AbstractConnection):
    __slots__ = "transport"

    transport: ZeroMQRPCTransport

    def __init__(self, transport):
        self.transport = transport

    async def recv_message(self) -> AsyncGenerator[Optional[RawHeaderBody], None]:
        assert not self.transport._closed
        assert self.transport._sock is not None
        while True:
            multipart_msg = await self.transport._sock.recv_multipart()
            try:
                *pre, zmsg_type, raw_header, raw_body = multipart_msg
                if zmsg_type == b"PING":
                    await self.transport._sock.send_multipart([
                        *pre,
                        b"PONG",
                        raw_header,
                        raw_body,
                    ])
                elif zmsg_type == b"UPPER":
                    if len(pre) > 0:
                        # server
                        peer_id = pre[0]
                        yield RawHeaderBody(raw_header, raw_body, peer_id)
                    else:
                        # client
                        yield RawHeaderBody(raw_header, raw_body, None)
                else:
                    # Ignore if the peer has sent a malformed multipart message.
                    log.debug(
                        "ZeroMQRPCConnection.recv_message(): "
                        "ignoring an invalid message from the peer..."
                    )
            except Exception as e:
                # ValueError may happen when there are garbage packets accepted by
                # the zmq socket.
                log.debug(
                    "ZeroMQRPCConnection.recv_message(): "
                    "exception caught in the recv loop, continuing...",
                    exc_info=e,
                )
                continue

    async def send_message(self, raw_msg: RawHeaderBody) -> None:
        assert not self.transport._closed
        assert self.transport._sock is not None
        peer_id = raw_msg.peer_id
        if peer_id is not None:
            # server
            await self.transport._sock.send_multipart([
                peer_id,
                b"UPPER",
                raw_msg.header,
                raw_msg.body,
            ])
        else:
            # client
            await self.transport._sock.send_multipart([
                b"UPPER",
                raw_msg.header,
                raw_msg.body,
            ])


class ZeroMQMonitorMixin:
    addr: ZeroMQAddress

    _monitor_sock: Optional[zmq.asyncio.Socket]

    async def _monitor(self) -> None:
        assert self._monitor_sock is not None
        log = logging.getLogger("callosum.lower.zeromq.monitor")
        try:
            while True:
                msg = await zmq.utils.monitor.recv_monitor_message(
                    self._monitor_sock
                )
                log.debug("monitor[%s] event: %r", self.addr, msg)
                # TODO: allow setting a custom callback
                if msg["event"] == zmq.EVENT_MONITOR_STOPPED:
                    break
        except Exception:
            log.exception("monitor[%s] unexpected error", self.addr)
        finally:
            self._monitor_sock.close()
            log.debug("monitor[%s] closed", self.addr)


class ZeroMQBaseBinder(ZeroMQMonitorMixin, AbstractBinder):
    __slots__ = (
        "transport",
        "addr",
        "_attach_monitor",
        "_monitor_sock",
        "_monitor_task",
        "_main_sock",
        "_zsock_opts",
    )

    socket_type: ClassVar[int] = 0

    transport: ZeroMQBaseTransport
    addr: ZeroMQAddress

    def __init__(
        self,
        transport: BaseTransport,
        addr: AbstractAddress,
        *,
        attach_monitor: bool = False,
        zsock_opts: Optional[Mapping[int, Any]] = None,
        **transport_opts,
    ) -> None:
        super().__init__(transport, addr)
        self._attach_monitor = attach_monitor
        self._zsock_opts = {**_default_zsock_opts, **(zsock_opts or {})}

    async def __aenter__(self):
        if not self.transport._closed:
            return ZeroMQRPCConnection(self.transport)
        server_sock = self.transport._zctx.socket(type(self).socket_type)
        await init_authenticator(self.transport.authenticator, server_sock)
        for key, value in self._zsock_opts.items():
            server_sock.setsockopt(key, value)
        if self._attach_monitor:
            monitor_addr = f"inproc://monitor-{secrets.token_hex(16)}"
            self._monitor_sock = server_sock.get_monitor_socket(addr=monitor_addr)
            self._monitor_task = asyncio.create_task(self._monitor())
        else:
            self._monitor_sock = None
            self._monitor_task = None
        server_sock.bind(self.addr.uri)
        self.transport._sock = server_sock
        self._main_sock = server_sock
        return ZeroMQRPCConnection(self.transport)

    async def __aexit__(self, exc_type, exc_obj, exc_tb):
        if self._monitor_task is not None:
            self._main_sock.disable_monitor()
            await self._monitor_task


class ZeroMQBaseConnector(ZeroMQMonitorMixin, AbstractConnector):
    __slots__ = (
        "transport",
        "addr",
        "_attach_monitor",
        "_monitor_sock",
        "_monitor_task",
        "_main_sock",
        "_zsock_opts",
    )

    _monitor_sock: Optional[zmq.asyncio.Socket]
    _main_sock: Optional[zmq.asyncio.Socket]
    _monitor_task: Optional[asyncio.Task]
    socket_type: ClassVar[int] = 0

    transport: ZeroMQBaseTransport
    addr: ZeroMQAddress

    def __init__(
        self,
        transport: BaseTransport,
        addr: AbstractAddress,
        *,
        attach_monitor: bool = False,
        handshake_timeout: float | None = None,
        zsock_opts: Optional[Mapping[int, Any]] = None,
        **transport_opts,
    ) -> None:
        super().__init__(transport, addr)
        self._attach_monitor = attach_monitor
        self._handshake_timeout = handshake_timeout
        self._zsock_opts = {**_default_zsock_opts, **(zsock_opts or {})}

    async def ping(self, ping_timeout: Optional[int] = None) -> bool:
        assert self._main_sock is not None
        sock: zmq.asyncio.Socket = self._main_sock
        await sock.send_multipart([b"PING", b"", b""])
        if await sock.poll(ping_timeout):
            try:
                response = await sock.recv_multipart(zmq.NOBLOCK)
            except zmq.Again:
                return False
            return response[0] == b"PONG"
        return False

    async def __aenter__(self) -> ZeroMQRPCConnection:
        if not self.transport._closed:
            return ZeroMQRPCConnection(self.transport)
        handshake_begin = time.perf_counter()
        client_sock = self.transport._zctx.socket(type(self).socket_type)
        await init_authenticator(self.transport.authenticator, client_sock)
        for key, value in self._zsock_opts.items():
            client_sock.setsockopt(key, value)
        if self._attach_monitor:
            monitor_addr = f"inproc://monitor-{secrets.token_hex(16)}"
            self._monitor_sock = client_sock.get_monitor_socket(addr=monitor_addr)
            self._monitor_task = asyncio.create_task(self._monitor())
        else:
            self._monitor_sock = None
            self._monitor_task = None
        client_sock.connect(self.addr.uri)
        self._main_sock = client_sock
        self.transport._sock = client_sock
        if not await self.ping(
            ping_timeout=int(self._handshake_timeout * 1000)
            if self._handshake_timeout is not None
            else 5000
        ):
            raise AuthenticationError
        handshake_done = time.perf_counter()
        log.debug(
            "ZeroMQ connector handshake latency: %.3f sec",
            handshake_done - handshake_begin,
        )
        return ZeroMQRPCConnection(self.transport)

    async def __aexit__(self, exc_type, exc_obj, exc_tb) -> Optional[bool]:
        assert self._main_sock is not None
        if self._monitor_task is not None:
            self._main_sock.disable_monitor()
            await self._monitor_task
        return None


class ZeroMQRouterBinder(ZeroMQBaseBinder):
    socket_type: ClassVar[int] = zmq.ROUTER
    transport: ZeroMQRPCTransport


class ZeroMQDealerConnector(ZeroMQBaseConnector):
    socket_type: ClassVar[int] = zmq.DEALER
    transport: ZeroMQRPCTransport


class ZeroMQPushBinder(ZeroMQBaseBinder):
    socket_type: ClassVar[int] = zmq.PUSH
    transport: ZeroMQDistributorTransport


class ZeroMQPullConnector(ZeroMQBaseConnector):
    socket_type: ClassVar[int] = zmq.PULL
    transport: ZeroMQDistributorTransport


class ZeroMQPubBinder(ZeroMQBaseBinder):
    socket_type: ClassVar[int] = zmq.PUB
    transport: ZeroMQBroadcastTransport


class ZeroMQSubConnector(ZeroMQBaseConnector):
    socket_type: ClassVar[int] = zmq.SUB
    transport: ZeroMQBroadcastTransport


class ZeroMQBaseTransport(BaseTransport):
    """
    Implementation for the ZeorMQ-backed transport.

    It keeps a single persistent connection over multiple connect/bind() calls.
    As the underlying PUSH/PULL sockets work asynchronously, this effectively
    achieves connection pooling reducing handshake overheads.
    """

    __slots__ = BaseTransport.__slots__ + (
        "_zctx",
        "_zsock_opts",
        "_zap_server",
        "_zap_task",
        "_sock",
    )

    _sock: Optional[zmq.asyncio.Socket]

    binder_cls: ClassVar[Type[ZeroMQBaseBinder]]
    connector_cls: ClassVar[Type[ZeroMQBaseConnector]]

    def __init__(
        self,
        authenticator: AbstractClientAuthenticator | AbstractServerAuthenticator,
        **kwargs,
    ) -> None:
        super().__init__(authenticator, **kwargs)
        self._zap_server = None
        self._zap_task = None
        self._zctx = zmq.asyncio.Context()
        match self.authenticator:
            case AbstractServerAuthenticator() as auth:
                self._zap_server = ZAPServer(self._zctx, auth)
                self._zap_task = asyncio.create_task(self._zap_server.serve())
            case _:
                pass
        # Keep sockets during the transport lifetime.
        self._sock = None

    @property
    def _closed(self) -> bool:
        if self._sock is not None:
            return self._sock.closed  # type: ignore
        return True

    async def close(self) -> None:
        if self._sock is not None:
            self._sock.close()
        if self._zap_task is not None and not self._zap_task.done():
            self._zap_task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await self._zap_task
        if self._zctx is not None:
            self._zctx.destroy(linger=50)


class ZeroMQRPCTransport(ZeroMQBaseTransport):
    binder_cls = ZeroMQRouterBinder
    connector_cls = ZeroMQDealerConnector


class ZeroMQDistributorTransport(ZeroMQBaseTransport):
    binder_cls = ZeroMQPushBinder
    connector_cls = ZeroMQPullConnector


class ZeroMQBroadcastTransport(ZeroMQBaseTransport):
    binder_cls = ZeroMQPubBinder
    connector_cls = ZeroMQSubConnector



from typing import Any, Protocol

from .channel import Peer
from .exceptions import RPCError, RPCInternalError, RPCUserError
from .message import RPCMessage


class FunctionHandler(Protocol):
    def __call__(self, request: RPCMessage, /) -> Any:  # noqa: E225
        ...


__all__ = (
    "RPCError",
    "RPCUserError",
    "RPCInternalError",
    "RPCMessage",
    "Peer",
    "FunctionHandler",
)



from __future__ import annotations

import asyncio
from typing import Any, AsyncGenerator, Mapping, Optional, Tuple, Union

import attrs
import redis
import redis.asyncio

from ..abc import RawHeaderBody
from . import (
    AbstractAddress,
    AbstractBinder,
    AbstractConnection,
    AbstractConnector,
    BaseTransport,
)
from .redis_common import redis_addr_to_url


@attrs.define(auto_attribs=True, slots=True)
class RedisStreamAddress(AbstractAddress):
    redis_server: Union[str, Tuple[str, int]]
    stream_key: str
    group: Optional[str] = None
    consumer: Optional[str] = None


class RPCRedisConnection(AbstractConnection):
    __slots__ = ("transport",)

    transport: RPCRedisTransport
    addr: RedisStreamAddress
    direction_keys: Optional[Tuple[str, str]]

    def __init__(
        self,
        transport: RPCRedisTransport,
        addr: RedisStreamAddress,
        direction_keys: Optional[Tuple[str, str]] = None,
    ):
        self.transport = transport
        self.addr = addr
        self.direction_keys = direction_keys

    async def recv_message(self) -> AsyncGenerator[Optional[RawHeaderBody], None]:
        assert self.transport._redis is not None
        assert self.addr.group is not None
        assert self.addr.consumer is not None
        if not self.direction_keys:
            stream_key = self.addr.stream_key
        else:
            stream_key = f"{self.addr.stream_key}.{self.direction_keys[0]}"

        try:
            per_key_fetch_list: list[Any] = []
            while not per_key_fetch_list:
                per_key_fetch_list = await self.transport._redis.xreadgroup(
                    self.addr.group,
                    self.addr.consumer,
                    {stream_key: ">"},
                    block=1000,
                )
            for fetch_info in per_key_fetch_list:
                if fetch_info[0].decode() != stream_key:
                    continue
                for item in fetch_info[1]:
                    item_id: bytes = item[0]
                    item_data: dict[bytes, bytes] = item[1]
                    try:
                        if b"meta" in item_data:
                            continue
                        yield RawHeaderBody(
                            item_data[b"hdr"],
                            item_data[b"msg"],
                            None,
                        )
                    finally:
                        await self.transport._redis.xack(
                            stream_key, self.addr.group, item_id
                        )
        except asyncio.CancelledError:
            raise
        except redis.asyncio.ConnectionError:
            yield None

    async def send_message(self, raw_msg: RawHeaderBody) -> None:
        assert self.transport._redis is not None
        if not self.direction_keys:
            stream_key = self.addr.stream_key
        else:
            stream_key = f"{self.addr.stream_key}.{self.direction_keys[1]}"
        await self.transport._redis.xadd(
            stream_key, {b"hdr": raw_msg[0], b"msg": raw_msg[1]}
        )


class RPCRedisBinder(AbstractBinder):
    """
    This class is binder for RPCRedisTransport.
    Usually, the RPC server is supposed to provide
    the connect address during Peer instantiation,
    which will result in the Peer class using
    the RPCRedisBinder for establishing
    the connection.
    """

    __slots__ = ("transport", "addr", "_addr_url")

    transport: RPCRedisTransport
    addr: RedisStreamAddress
    _addr_url: str

    async def __aenter__(self):
        assert self.addr.group is not None
        assert self.addr.consumer is not None
        self._addr_url = redis_addr_to_url(self.addr.redis_server)
        self.transport._redis = await redis.asyncio.from_url(
            self._addr_url, **self.transport._redis_opts
        )
        key = f"{self.addr.stream_key}.bind"
        await self.transport._redis.xadd(key, {b"meta": b"create-stream"})
        groups = await self.transport._redis.xinfo_groups(key)
        print(groups)
        if not any(map(lambda g: g["name"] == self.addr.group.encode(), groups)):
            await self.transport._redis.xgroup_create(
                key,
                self.addr.group,
                mkstream=True,
            )
        return RPCRedisConnection(self.transport, self.addr, ("bind", "conn"))

    async def __aexit__(self, exc_type, exc_obj, exc_tb):
        assert self.addr.group is not None
        assert self.addr.consumer is not None
        # we need to create a new Redis connection for cleanup
        # because self.transport._redis gets corrupted upon
        # cancellation of Peer._recv_loop() task.
        _redis = await redis.asyncio.from_url(
            self._addr_url, **self.transport._redis_opts
        )
        try:
            await asyncio.shield(
                _redis.xgroup_delconsumer(
                    f"{self.addr.stream_key}.bind",
                    self.addr.group,
                    self.addr.consumer,
                )
            )
        finally:
            await _redis.close()


class RPCRedisConnector(AbstractConnector):
    """
    This class is connector for RPCRedisTransport.
    Usually, the RPC invoker (client) is supposed to provide
    the connect address during Peer instantiation,
    which will result in the Peer class using
    the RPCRedisConnector for establishing
    the connection.
    """

    __slots__ = ("transport", "addr", "_addr_url")

    transport: RPCRedisTransport
    addr: RedisStreamAddress

    async def __aenter__(self):
        assert self.addr.group is not None
        assert self.addr.consumer is not None
        self._addr_url = redis_addr_to_url(self.addr.redis_server)
        self.transport._redis = redis.asyncio.from_url(
            self._addr_url, **self.transport._redis_opts
        )
        key = f"{self.addr.stream_key}.conn"
        await self.transport._redis.xadd(key, {b"meta": b"create-stream"})
        groups = await self.transport._redis.xinfo_groups(key)
        if not any(map(lambda g: g["name"] == self.addr.group.encode(), groups)):
            await self.transport._redis.xgroup_create(
                key,
                self.addr.group,
                mkstream=True,
            )
        return RPCRedisConnection(self.transport, self.addr, ("conn", "bind"))

    async def __aexit__(self, exc_type, exc_obj, exc_tb):
        assert self.addr.group is not None
        assert self.addr.consumer is not None
        # we need to create a new Redis connection for cleanup
        # because self.transport._redis gets corrupted upon
        # cancellation of Peer._recv_loop() task.
        _redis = await redis.asyncio.from_url(
            self._addr_url, **self.transport._redis_opts
        )
        try:
            await asyncio.shield(
                _redis.xgroup_delconsumer(
                    f"{self.addr.stream_key}.conn",
                    self.addr.group,
                    self.addr.consumer,
                )
            )
        finally:
            await _redis.close()


class RPCRedisTransport(BaseTransport):
    """
    Implementation for bidirectional transport backend by Redis Streams.
    """

    __slots__ = BaseTransport.__slots__ + (
        "_redis_opts",
        "_redis",
    )

    _redis_opts: Mapping[str, Any]
    _redis: Optional[redis.asyncio.Redis]

    binder_cls = RPCRedisBinder
    connector_cls = RPCRedisConnector

    def __init__(
        self,
        authenticator,
        **kwargs,
    ) -> None:
        super().__init__(authenticator, **kwargs)
        self._redis_opts = self.transport_opts.get("redis_opts", {})
        self._redis = None

    async def close(self) -> None:
        if self._redis is not None:
            await self._redis.close()



from __future__ import annotations

import asyncio
import functools
import logging
from contextlib import aclosing
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Mapping,
    MutableMapping,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
)

import attrs

from ..abc import (
    AbstractChannel,
    AbstractDeserializer,
    AbstractSerializer,
    QueueSentinel,
    TaskSentinel,
)
from ..auth import AbstractClientAuthenticator, AbstractServerAuthenticator
from ..lower import (
    AbstractAddress,
    AbstractBinder,
    AbstractConnection,
    AbstractConnector,
    BaseTransport,
)
from ..ordering import (
    SEQ_BITS,
    AbstractAsyncScheduler,
    AsyncResolver,
    ExitOrderedAsyncScheduler,
)
from .exceptions import RPCInternalError, RPCUserError
from .message import RPCMessage, RPCMessageTypes
from .types import RequestId

if TYPE_CHECKING:
    from . import FunctionHandler

log = logging.getLogger(__spec__.name)  # type: ignore[name-defined]


class Peer(AbstractChannel):
    """
    Represents a bidirectional connection where both sides can invoke each
    other.

    In Callosum, there is no fixed server or client for a connection.
    Once the connection is established, each peer can become both
    RPC client and RPC server.
    """

    _connection: Optional[AbstractConnection]
    _deserializer: AbstractDeserializer
    _serializer: AbstractSerializer
    _func_registry: MutableMapping[str, FunctionHandler]
    _outgoing_queue: asyncio.Queue[Union[QueueSentinel, RPCMessage]]
    _recv_task: Optional[asyncio.Task]
    _send_task: Optional[asyncio.Task]
    _opener: Optional[Union[AbstractBinder, AbstractConnector]]

    # The mapping from (peer ID, client request ID) -> server request ID
    _req_idmap: Dict[Tuple[Any, RequestId], RequestId]

    _log: logging.Logger
    _debug_rpc: bool

    def __init__(
        self,
        *,
        deserializer: AbstractDeserializer,
        serializer: AbstractSerializer,
        connect: Optional[AbstractAddress] = None,
        bind: Optional[AbstractAddress] = None,
        transport: Optional[Type[BaseTransport]] = None,
        authenticator: Optional[
            AbstractClientAuthenticator | AbstractServerAuthenticator
        ] = None,
        transport_opts: Optional[Mapping[str, Any]] = None,
        scheduler: Optional[AbstractAsyncScheduler] = None,
        compress: bool = True,
        max_body_size: int = 10 * (2**20),  # 10 MiBytes
        max_concurrency: int = 100,
        execute_timeout: Optional[float] = None,
        invoke_timeout: Optional[float] = None,
        debug_rpc: bool = False,
    ) -> None:
        if connect is None and bind is None:
            raise ValueError("You must specify either the connect or bind address.")
        self._connect = connect
        self._bind = bind
        self._opener = None
        self._connection = None
        self._compress = compress
        self._deserializer = deserializer
        self._serializer = serializer
        self._max_concurrency = max_concurrency
        self._exec_timeout = execute_timeout
        self._invoke_timeout = invoke_timeout

        self._scheduler = None
        if transport is None:
            raise ValueError("You must provide a transport class.")
        self._transport = transport(
            authenticator=authenticator,
            transport_opts=transport_opts or {},
        )
        self._func_registry = {}

        self._client_seq_id = 0
        self._server_seq_id = 0
        self._req_idmap = {}

        # incoming queues
        self._invocation_resolver = AsyncResolver()
        if scheduler is None:
            scheduler = ExitOrderedAsyncScheduler()
        self._func_scheduler = scheduler

        # there is only one outgoing queue
        self._outgoing_queue = asyncio.Queue()
        self._recv_task = None
        self._send_task = None

        self._log = logging.getLogger(__name__ + ".Peer")
        self._debug_rpc = debug_rpc

    def handle_function(self, method: str, handler: FunctionHandler) -> None:
        self._func_registry[method] = handler

    def unhandle_function(self, method: str) -> None:
        del self._func_registry[method]

    def _lookup_func(self, method: str) -> FunctionHandler:
        return self._func_registry[method]

    async def _recv_loop(self) -> None:
        """
        Receive requests and schedule the request handlers.
        """
        if self._connection is None:
            raise RuntimeError("consumer is not opened yet.")
        func_tasks: Set[asyncio.Task] = set()
        try:
            async with aclosing(self._connection.recv_message()) as agen:
                async for raw_msg in agen:
                    # TODO: flow-control in transports or peer queues?
                    if raw_msg is None:
                        return
                    request = RPCMessage.decode(raw_msg, self._deserializer)
                    client_request_id = request.request_id
                    server_request_id: Optional[RequestId]
                    if request.msgtype == RPCMessageTypes.FUNCTION:
                        server_seq_id = self._next_server_seq_id()
                        server_request_id = (
                            client_request_id[0],
                            client_request_id[1],
                            server_seq_id,
                        )
                        self._req_idmap[(request.peer_id, client_request_id)] = (
                            server_request_id
                        )
                        func_handler = self._lookup_func(request.method)
                        task = asyncio.create_task(
                            self._func_task(
                                server_request_id,
                                request,
                                func_handler,
                            )
                        )
                        func_tasks.add(task)
                        task.add_done_callback(func_tasks.discard)
                        task.add_done_callback(
                            lambda task: self._req_idmap.pop(
                                (request.peer_id, client_request_id),
                                None,
                            )
                        )
                        await asyncio.sleep(0)
                    elif request.msgtype == RPCMessageTypes.CANCEL:
                        server_request_id = self._req_idmap.pop(
                            (request.peer_id, client_request_id),
                            None,
                        )
                        if server_request_id is None:
                            continue
                        await self._func_scheduler.cancel(server_request_id)
                    elif request.msgtype in (
                        RPCMessageTypes.RESULT,
                        RPCMessageTypes.FAILURE,
                        RPCMessageTypes.CANCELLED,
                        RPCMessageTypes.ERROR,
                    ):
                        self._invocation_resolver.resolve(
                            client_request_id,
                            request,
                        )
                        await asyncio.sleep(0)
        except asyncio.CancelledError:
            pending_tasks = []
            if func_tasks:
                for task in func_tasks:
                    if not task.done():
                        task.cancel()
                        pending_tasks.append(task)
                await asyncio.wait(pending_tasks)
            await asyncio.sleep(0)
        except Exception:
            log.exception("unexpected error")

    async def _send_loop(self) -> None:
        """
        Fetches and sends out the completed task responses.
        """
        if self._connection is None:
            raise RuntimeError("consumer is not opened yet.")
        while True:
            try:
                msg = await self._outgoing_queue.get()
                try:
                    if msg is QueueSentinel.CLOSED:
                        break
                    await self._connection.send_message(msg.encode(self._serializer))
                finally:
                    self._outgoing_queue.task_done()
            except asyncio.CancelledError:
                break
            except Exception:
                log.exception("unexpected error")

    async def __aenter__(self) -> Peer:
        _opener: Union[AbstractBinder, AbstractConnector]
        if self._connect:
            _opener = functools.partial(self._transport.connect, self._connect)()
        elif self._bind:
            _opener = functools.partial(self._transport.bind, self._bind)()
        else:
            raise RuntimeError("Misconfigured opener")
        self._opener = _opener
        self._connection = await _opener.__aenter__()
        # NOTE: if we change the order of the following 2 lines of code,
        # then there will be error after "flushall" redis.
        self._send_task = asyncio.create_task(self._send_loop())
        self._recv_task = asyncio.create_task(self._recv_loop())
        return self

    async def __aexit__(self, *exc_info) -> None:
        if self._send_task is not None:
            await self._outgoing_queue.put(QueueSentinel.CLOSED)
            await self._send_task
        if self._recv_task is not None:
            # TODO: pass exception description, e.g. during invoke timeout
            self._recv_task.cancel()
            await self._recv_task
        if self._opener is not None:
            await self._opener.__aexit__(*exc_info)
        if self._transport is not None:
            await self._transport.close()
        # TODO: add proper cleanup for awaiting on
        # finishing of the "listen" coroutine's spawned tasks

    def _next_client_seq_id(self) -> int:
        current = self._client_seq_id
        self._client_seq_id = (self._client_seq_id + 1) % (2**SEQ_BITS)
        return current

    def _next_server_seq_id(self) -> int:
        current = self._server_seq_id
        self._server_seq_id = (self._server_seq_id + 1) % (2**SEQ_BITS)
        return current

    async def _func_task(
        self,
        server_request_id: Tuple[str, str, int],
        request: RPCMessage,
        handler: FunctionHandler,
    ) -> None:
        response = RPCMessage.cancelled(request)
        try:
            await self._func_scheduler.schedule(server_request_id, handler(request))
            try:
                result = await self._func_scheduler.get_fut(server_request_id)
                if result is TaskSentinel.CANCELLED:
                    response = RPCMessage.cancelled(request)
            except asyncio.CancelledError:
                raise
            except Exception:
                # exception from user handler => failure
                if self._debug_rpc:
                    self._log.exception("RPC user error")
                response = RPCMessage.failure(request)
            else:
                response = RPCMessage.result(request, result)
        except asyncio.CancelledError:
            response = RPCMessage.cancelled(request)
            raise
        except Exception:
            if self._debug_rpc:
                self._log.exception("RPC internal error")
            # exception from our parts => error
            response = RPCMessage.error(request)
        finally:
            self._func_scheduler.cleanup(server_request_id)
            await self._outgoing_queue.put(response)

    async def invoke(
        self, method: str, body, *, order_key=None, invoke_timeout=None
    ):
        """
        Invoke a remote function via the transport connection.
        """
        if invoke_timeout is None:
            invoke_timeout = self._invoke_timeout
        if order_key is None:
            order_key = b""
        client_seq_id = self._next_client_seq_id()
        try:
            request: RPCMessage
            server_cancelled = False
            async with asyncio.timeout(invoke_timeout):
                if callable(body):
                    # The user is using an upper-layer adaptor.
                    async with aclosing(body()) as agen:
                        request = RPCMessage(
                            None,
                            RPCMessageTypes.FUNCTION,
                            method,
                            order_key,
                            client_seq_id,
                            None,
                            await agen.asend(None),
                        )
                        await self._outgoing_queue.put(request)
                        response = await self._invocation_resolver.wait(
                            request.request_id
                        )
                        upper_result = await agen.asend(response.body)
                        try:
                            await agen.asend(None)
                        except StopAsyncIteration:
                            pass
                else:
                    request = RPCMessage(
                        None,
                        RPCMessageTypes.FUNCTION,
                        method,
                        order_key,
                        client_seq_id,
                        None,
                        body,
                    )
                    await self._outgoing_queue.put(request)
                    response = await self._invocation_resolver.wait(
                        request.request_id
                    )
                    upper_result = response.body
            if response.msgtype == RPCMessageTypes.RESULT:
                pass
            elif response.msgtype == RPCMessageTypes.CANCELLED:
                server_cancelled = True
                raise asyncio.CancelledError
            elif response.msgtype == RPCMessageTypes.FAILURE:
                raise RPCUserError(*attrs.astuple(response.metadata))
            elif response.msgtype == RPCMessageTypes.ERROR:
                raise RPCInternalError(*attrs.astuple(response.metadata))
            return upper_result
        except (asyncio.TimeoutError, asyncio.CancelledError):
            # propagate cancellation to the connected peer
            if not server_cancelled:
                cancel_request = RPCMessage.cancel(request)
                await self._outgoing_queue.put(cancel_request)
            raise
        except Exception:
            raise



import json

import pytest

from callosum.lower.zeromq import ZeroMQAddress, ZeroMQRPCTransport
from callosum.rpc import Peer


@pytest.mark.asyncio
async def test_init():
    p = Peer(
        connect=ZeroMQAddress("tcp://127.0.0.1:5000"),
        serializer=json.dumps,
        deserializer=json.loads,
        transport=ZeroMQRPCTransport,
    )

    assert p._connect.uri == "tcp://127.0.0.1:5000"
    assert p._max_concurrency > 0
    assert p._invoke_timeout is None or p._invoke_timeout > 0
    assert p._exec_timeout is None or p._exec_timeout > 0
    assert len(p._func_registry) == 0


@pytest.mark.asyncio
async def test_func_registry():
    p = Peer(
        connect=ZeroMQAddress("tcp://127.0.0.1:5000"),
        serializer=json.dumps,
        deserializer=json.loads,
        transport=ZeroMQRPCTransport,
    )

    def dummy():
        pass

    assert "dummy" not in p._func_registry
    p.handle_function("dummy", dummy)
    assert "dummy" in p._func_registry
    assert p._func_registry["dummy"] is dummy



import argparse
import re
import subprocess
import sys
from pathlib import Path


def get_current_tag():
    p = subprocess.run(
        ["git", "describe", "--abbrev=0", "--tags"], capture_output=True
    )
    revision = p.stdout.decode().strip()
    return revision


def get_prev_tag(current_tag):
    p = subprocess.run(
        ["git", "describe", "--abbrev=0", "--tags", f"{current_tag}^"],
        capture_output=True,
    )
    tag = p.stdout.decode().strip()
    return tag


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--draft",
        action="store_true",
        default=False,
        help="Do not write to file but only print the expected output",
    )
    args = parser.parse_args()

    current_tag = get_current_tag()
    prev_tag = get_prev_tag(current_tag)
    commitlog_url = (
        f"https://github.com/lablup/callosum/compare/{prev_tag}...{current_tag}"
    )
    changelog_url = (
        f"https://github.com/lablup/callosum/blob/{current_tag}/CHANGELOG.md"
    )

    print(f"Making release notes for {current_tag} ...", file=sys.stderr)

    input_path = Path("./CHANGES.md")
    output_path = Path("./CHANGELOG_RELEASE.md")
    try:
        version = current_tag
        input_text = input_path.read_text()
        m = re.search(
            rf"(?:^|\n)## v?{re.escape(version)}(?:[^\n]*)?\n(.*?)(?:\n## |$)",
            input_text,
            re.S,
        )
        if m is not None:
            content = m.group(1).strip()
            content += "\n"
            content += (
                f"\n### Full Changelog"
                f"\nCheck out [the full changelog]({changelog_url}) "
                f"until this release ({current_tag}).\n"
            )
            content += (
                f"\n### Full Commit Logs"
                f"\nCheck out [the full commit logs]({commitlog_url}) "
                f"between release ({prev_tag}) and ({current_tag}).\n"
            )
            if not args.draft:
                output_path.write_text(content)
            print("--------")
            print(content)
            print("--------")
            print(
                "Successfully extracted the latest changelog to CHANGELOG_RELEASE.md",
                file=sys.stderr,
            )
        else:
            print(
                "::error ::Could not extract the latest changelog from CHANGELOG.md",
                file=sys.stderr,
            )
            sys.exit(1)
    except IOError as e:
        print(f"::error ::Could read or write from file: {e!r}", file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    main()



from __future__ import annotations

import abc
import asyncio
import functools
import heapq
import logging
from collections import defaultdict
from typing import Any, Awaitable, Dict, Final, List, Optional, Union

import attrs

from .abc import TaskSentinel
from .serial import serial_lt

SEQ_BITS: Final = 32


def _resolve_future(request_id, fut, result, log) -> None:
    if fut is None:
        log.warning("resolved unknown request: %r", request_id)
        return
    if fut.done():
        if fut.cancelled():
            log.debug("resolved cancelled request: %r", request_id)
        elif fut.exception() is not None:
            log.debug("resolved errored request: %r", request_id)
        return
    if isinstance(result, BaseException):
        fut.set_exception(result)
    else:
        fut.set_result(result)


class AsyncResolver:
    __slots__ = ("_log", "_futures")

    _futures: Dict[_SeqItem, asyncio.Future]

    def __init__(self) -> None:
        self._log = logging.getLogger(__name__ + ".AsyncResolver")
        self._futures = {}

    def wait(self, request_id) -> asyncio.Future:
        if request_id in self._futures:
            raise RuntimeError("duplicate request: %r", request_id)
        loop = asyncio.get_running_loop()
        fut = loop.create_future()
        self._futures[request_id] = fut
        return fut

    def cancel(self, request_id) -> None:
        fut = self._futures.pop(request_id, None)
        if fut is not None and not fut.done():
            fut.cancel()

    def resolve(self, request_id, result) -> None:
        fut = self._futures.pop(request_id, None)
        _resolve_future(request_id, fut, result, self._log)

    async def cleanup(self, request_id) -> None:
        pass


class AbstractAsyncScheduler(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def schedule(self, request_id, coro) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def get_fut(self, request_id) -> Awaitable[Union[TaskSentinel, Any]]:
        raise NotImplementedError

    @abc.abstractmethod
    async def cancel(self, request_id) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def cleanup(self, request_id) -> None:
        raise NotImplementedError


@functools.total_ordering
@attrs.define(frozen=True, slots=True, eq=False, order=False)
class _SeqItem:
    method: str
    seq: int
    ev: Optional[asyncio.Event] = None

    def __lt__(self, other: _SeqItem) -> bool:
        return serial_lt(self.seq, other.seq, SEQ_BITS)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, _SeqItem):
            raise TypeError("cannot compare equality with non-SeqItem object", other)
        return self.seq == other.seq


class KeySerializedAsyncScheduler(AbstractAsyncScheduler):
    """
    A scheduler which ensures serialized returning of results within the given
    ordering key while it allows overlapped execution.
    """

    __slots__ = ("_log", "_futures", "_tasks", "_pending")

    _futures: Dict[_SeqItem, asyncio.Future]
    _tasks: Dict[_SeqItem, asyncio.Task]
    _pending: Dict[bytes, List[_SeqItem]]

    def __init__(self) -> None:
        self._log = logging.getLogger(__name__ + ".KeySerializedAsyncScheduler")
        self._futures = {}
        self._tasks = {}
        self._pending = defaultdict(list)

    async def schedule(self, request_id, coro) -> None:
        method, okey, seq = request_id
        loop = asyncio.get_running_loop()
        fut = loop.create_future()
        self._futures[request_id] = fut
        ev = asyncio.Event()
        heapq.heappush(self._pending[okey], _SeqItem(method, seq, ev))

        while True:
            assert len(self._pending[okey]) > 0
            head = self._pending[okey][0]
            if head.seq == seq:
                break
            # Wait until the head item finishes.
            assert head.ev is not None
            await head.ev.wait()

        task = asyncio.create_task(coro)
        self._tasks[request_id] = task

        def cb(seq_item, rqst_id, fut, task):
            _, okey, _ = rqst_id
            try:
                s = heapq.heappop(self._pending[okey])
                assert s == seq_item
                if task.cancelled():
                    # already removed from the pending heap
                    fut.cancel()
                else:
                    if task.exception() is not None:
                        _resolve_future(rqst_id, fut, task.exception(), self._log)
                    else:
                        _resolve_future(rqst_id, fut, task.result(), self._log)
            finally:
                seq_item.ev.set()

        task.add_done_callback(functools.partial(cb, head, request_id, fut))

    def get_fut(self, request_id) -> Awaitable[Union[TaskSentinel, Any]]:
        return self._futures[request_id]

    def cleanup(self, request_id) -> None:
        self._futures.pop(request_id, None)
        okey = request_id[1]
        if okey in self._pending:
            for seq_item in self._pending[okey]:
                if seq_item.seq == request_id[2]:
                    self._pending[okey].remove(seq_item)
                    break
        if len(self._pending[okey]) == 0:
            del self._pending[okey]
        else:
            heapq.heapify(self._pending[okey])
        self._tasks.pop(request_id, None)

    async def cancel(self, request_id) -> None:
        if request_id in self._futures:
            task = self._tasks.get(request_id, None)
            if task is None:
                return
            if not task.done():
                task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                pass
            # The task callback will notify the scheduler caller (upper).
            # The scheduler caller (upper) will call cleanup.
        else:
            self._log.warning(
                "cancellation of unknown or " "not sent yet request: %r", request_id
            )


class ExitOrderedAsyncScheduler(AbstractAsyncScheduler):
    """
    A scheduler which let the asyncio's event loop do its own scheduling.
    """

    __slots__ = ("_log", "_tasks")

    _tasks: Dict[_SeqItem, Union[TaskSentinel, asyncio.Task]]

    def __init__(self):
        self._log = logging.getLogger(__name__ + ".ExitOrderedAsyncScheduler")
        self._tasks = {}

    async def schedule(self, request_id, coro) -> None:
        task = asyncio.create_task(coro)
        if request_id in self._tasks:
            raise RuntimeError("duplicate request: %r", request_id)
        self._tasks[request_id] = task

    async def get_fut(self, request_id) -> Union[TaskSentinel, Any]:
        task = self._tasks.pop(request_id, TaskSentinel.CANCELLED)
        if isinstance(task, TaskSentinel):
            return task
        return await task

    def cleanup(self, request_id) -> None:
        self._tasks.pop(request_id, None)

    async def cancel(self, request_id) -> None:
        task = self._tasks.get(request_id, TaskSentinel.CANCELLED)
        if task is TaskSentinel.CANCELLED:
            return
        if not task.done():
            task.cancel()
        self._tasks[request_id] = TaskSentinel.CANCELLED
        await task



import asyncio
import json
import logging
import os
import random
import secrets

from callosum.lower.rpc_redis import RedisStreamAddress, RPCRedisTransport
from callosum.rpc import Peer


async def call():
    redis_host = os.environ.get("REDIS_HOST", "127.0.0.1")
    redis_port = int(os.environ.get("REDIS_PORT", "6379"))
    peer = Peer(
        connect=RedisStreamAddress(
            f"redis://{redis_host}:{redis_port}",
            "myservice",
            "client-group",
            "server1",
        ),
        transport=RPCRedisTransport,
        serializer=json.dumps,
        deserializer=json.loads,
        invoke_timeout=3.0,
    )
    async with peer:
        response = await peer.invoke(
            "echo",
            {
                "sent": secrets.token_hex(16),
            },
        )
        print(f"echoed {response['received']}")
        response = await peer.invoke(
            "echo",
            {
                "sent": secrets.token_hex(16),
            },
        )
        print(f"echoed {response['received']}")
        a = random.randint(1, 10)
        b = random.randint(10, 20)
        response = await peer.invoke(
            "add",
            {
                "a": a,
                "b": b,
            },
        )
        print(f"{a} + {b} = {response['result']}")


if __name__ == "__main__":
    logging.basicConfig(
        format="%(asctime)s [%(levelname)s] %(name)s: %(message)s",
        level=logging.INFO,
    )
    log = logging.getLogger()
    asyncio.run(call())



from __future__ import annotations

import abc
import enum
from typing import Any, NamedTuple, Optional, Protocol


class RawHeaderBody(NamedTuple):
    header: bytes
    body: bytes
    peer_id: Optional[bytes]


class QueueSentinel(enum.Enum):
    """
    A category of special singleton objects that represents
    control-plane events in data-plane RX/TX queues.
    """

    CLOSED = 1


class TaskSentinel(enum.Enum):
    """
    A category of special singleton objects that represents
    special status of asyncio tasks.
    """

    CANCELLED = 1


class AbstractSerializer(Protocol):
    def __call__(self, obj: Optional[Any], /) -> bytes:  # noqa: E225
        ...


class AbstractDeserializer(Protocol):
    def __call__(self, data: bytes, /) -> Optional[Any]:  # noqa: E225
        ...


class AbstractMessage(metaclass=abc.ABCMeta):
    @classmethod
    @abc.abstractmethod
    def decode(
        cls, raw_msg: RawHeaderBody, deserializer: AbstractDeserializer
    ) -> AbstractMessage:
        """
        Decodes the message and applies deserializer to the body.
        Returns an instance of inheriting message class.

        Args:
            deserializer: Body deserializer.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def encode(self, serializer: AbstractSerializer) -> RawHeaderBody:
        """
        Encodes the message and applies serializer to body.

        Args:
            serializer: Body serializer.
        """
        raise NotImplementedError


class AbstractChannel(metaclass=abc.ABCMeta):
    async def __aenter__(self) -> AbstractChannel:
        return self

    async def __aexit__(self, *exc_info) -> None:
        pass



from __future__ import annotations

import abc
from typing import Optional

import attrs
import zmq


@attrs.define(frozen=True, slots=True)
class AuthResult:
    success: bool
    user_id: Optional[str] = None


@attrs.define(frozen=True, slots=True)
class Identity:
    domain: str
    private_key: bytes


@attrs.define(frozen=True, slots=True)
class Credential:
    domain: str
    public_key: bytes


def create_keypair():
    """
    Generate a new CURVE-25519 public-private keypair.
    """
    # NOTE: currently we rely on zmq for convenience, but we may use libnacl directly
    #       if we want to isolate this module from zmq dependency.
    public_key, private_key = zmq.curve_keypair()
    return public_key, private_key


class AbstractServerAuthenticator(metaclass=abc.ABCMeta):
    """
    Users of Callosum should subclass this to implement custom authentication.

    Though `lower.zeromq` uses the keypair to encrypt the traffic as well as
    authenticate the peer sockets, but this is not a mandatory requirement for
    transport implementations.  A transport may simply use its own network-level
    encryption and/or authentication scheme while leaving this authenticator as an
    application-level identity management scheme.
    """

    @abc.abstractmethod
    async def server_identity(self) -> Identity:
        """
        Return the identity of the server.
        Only used by the binder.
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def check_client(self, creds: Credential) -> AuthResult:
        """
        Check if the given domain and client public key is a valid one or not.
        Only used by the binder.
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def server_public_key(self) -> bytes:
        """
        Return the public key of the server.
        Only used by the connector.
        """
        raise NotImplementedError


class AbstractClientAuthenticator(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def server_public_key(self) -> bytes:
        """
        Return the public key of the server.
        Only used by the connector.
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def client_identity(self) -> Identity:
        """
        Return the identity of the client.
        Only used by the connector.
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def client_public_key(self) -> bytes:
        """
        Return the public key of the client.
        Only used by the connector.
        """
        raise NotImplementedError



import asyncio
import json
import random
from typing import Any, Awaitable, Callable, List, Mapping, cast

import pytest

from callosum.lower.zeromq import ZeroMQAddress, ZeroMQRPCTransport
from callosum.ordering import (
    AbstractAsyncScheduler,
    ExitOrderedAsyncScheduler,
    KeySerializedAsyncScheduler,
)
from callosum.rpc.channel import Peer
from callosum.rpc.exceptions import RPCUserError
from callosum.rpc.message import (
    ErrorMetadata,
    NullMetadata,
    RPCMessage,
    RPCMessageTypes,
)


def test_metadata_serialization():
    orig = NullMetadata()
    data = orig.encode()
    out = NullMetadata.decode(data)
    assert out == orig

    orig = ErrorMetadata("MyError", "MyError()", "this is a long traceback")
    data = orig.encode()
    out = ErrorMetadata.decode(data)
    assert out == orig


def test_rpcmessage_exception_serialization():
    request = RPCMessage(
        peer_id=None,
        msgtype=RPCMessageTypes.FUNCTION,
        method="dummy_function",
        order_key="x",
        client_seq_id=1000,
        metadata=NullMetadata(),
        body=b"{}",
    )
    try:
        raise ZeroDivisionError("oops")
    except Exception:
        failure_msg = RPCMessage.failure(request)
        assert failure_msg.msgtype == RPCMessageTypes.FAILURE
        data = failure_msg.encode(json.dumps)
        decoded_failure_msg = RPCMessage.decode(data, json.loads)
        assert decoded_failure_msg == failure_msg


async def dummy_server(
    scheduler: AbstractAsyncScheduler,
    func: Callable[[RPCMessage], Any],
    done_event: asyncio.Event,
) -> None:
    server = Peer(
        bind=ZeroMQAddress("tcp://127.0.0.1:5020"),
        transport=ZeroMQRPCTransport,
        scheduler=scheduler,
        serializer=lambda o: json.dumps(o).encode("utf8"),
        deserializer=lambda b: json.loads(b),
    )
    server.handle_function("func", func)
    async with server:
        await done_event.wait()


async def dummy_client(
    requester: Callable[[Peer], Awaitable[None]],
    done_event: asyncio.Event,
) -> None:
    client = Peer(
        connect=ZeroMQAddress("tcp://localhost:5020"),
        transport=ZeroMQRPCTransport,
        serializer=lambda o: json.dumps(o).encode("utf8"),
        deserializer=lambda b: json.loads(b),
    )
    async with client:
        await requester(client)
    done_event.set()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "scheduler_cls",
    [ExitOrderedAsyncScheduler, KeySerializedAsyncScheduler],
)
async def test_messaging(scheduler_cls) -> None:
    done = asyncio.Event()
    total = 50
    call_results: List[int | BaseException] = []
    return_results: List[int] = []
    # a list of events to make fucntions to return in the reversed order
    order_events = [asyncio.Event() for _ in range(total)]

    async def func(request: RPCMessage) -> int:
        # waits a short delay so that it return in the reversed order
        body = cast(Mapping[str, int], request.body)
        total = body["total"]
        idx = body["idx"]
        if idx == 0:
            await asyncio.sleep(0.01)
            for j in range(total, 0, -1):
                order_events[j - 1].set()
                await asyncio.sleep(0.01)  # let the next waiter proceed
        await order_events[idx].wait()
        return idx

    async def requester(client) -> None:
        async def _do_request(idx: int) -> int:
            ret = await client.invoke(
                "func",
                {
                    "total": total,
                    "idx": idx,
                },
                order_key="mykey",
            )
            return_results.append(ret)
            return ret

        tasks = []
        for idx in range(total):
            tasks.append(asyncio.create_task(_do_request(idx)))
        call_results.extend(await asyncio.gather(*tasks, return_exceptions=True))

    scheduler = scheduler_cls()
    server_task = asyncio.create_task(dummy_server(scheduler, func, done))
    client_task = asyncio.create_task(dummy_client(requester, done))
    await asyncio.wait([server_task, client_task])
    if isinstance(scheduler, KeySerializedAsyncScheduler):
        # check memory leak
        assert len(scheduler._pending) == 0
        assert len(scheduler._tasks) == 0
        assert len(scheduler._futures) == 0
    elif isinstance(scheduler, ExitOrderedAsyncScheduler):
        assert len(scheduler._tasks) == 0

    for idx in range(total):
        assert call_results[idx] == idx
        if isinstance(scheduler, KeySerializedAsyncScheduler):
            assert return_results[idx] == idx
        else:
            assert return_results[idx] == total - idx - 1


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "scheduler_cls",
    [ExitOrderedAsyncScheduler, KeySerializedAsyncScheduler],
)
async def test_messaging_server_cancellation(scheduler_cls) -> None:
    done = asyncio.Event()
    total = 200
    cancel_idxs = [False] * 130 + [True] * 70
    random.shuffle(cancel_idxs)
    call_results: List[int | BaseException] = []
    return_results: List[int] = []
    # a list of events to make fucntions to return in the reversed order
    order_events = [asyncio.Event() for _ in range(total)]

    async def func(request: RPCMessage) -> int:
        # waits a short delay so that it return in the reversed order
        body = cast(Mapping[str, int], request.body)
        total = body["total"]
        idx = body["idx"]
        if idx == 0:
            await asyncio.sleep(0.01)
            for j in range(total, 0, -1):
                order_events[j - 1].set()
                await asyncio.sleep(0.01)  # let the next waiter proceed
        await order_events[idx].wait()
        if cancel_idxs[idx]:
            raise asyncio.CancelledError
        return idx

    async def requester(client) -> None:
        async def _do_request(idx: int) -> int:
            try:
                ret = await client.invoke(
                    "func",
                    {
                        "total": total,
                        "idx": idx,
                    },
                    order_key="mykey",
                )
            except asyncio.CancelledError:
                return_results.append(-1)
                raise
            else:
                return_results.append(ret)
            return ret

        tasks = []
        for idx in range(total):
            tasks.append(asyncio.create_task(_do_request(idx)))
        call_results.extend(await asyncio.gather(*tasks, return_exceptions=True))

    scheduler = scheduler_cls()
    server_task = asyncio.create_task(dummy_server(scheduler, func, done))
    client_task = asyncio.create_task(dummy_client(requester, done))
    await asyncio.wait([server_task, client_task])

    if isinstance(scheduler, KeySerializedAsyncScheduler):
        # check memory leak
        assert len(scheduler._pending) == 0
        assert len(scheduler._tasks) == 0
        assert len(scheduler._futures) == 0
    elif isinstance(scheduler, ExitOrderedAsyncScheduler):
        assert len(scheduler._tasks) == 0

    for idx in range(total):
        if cancel_idxs[idx]:
            assert isinstance(call_results[idx], asyncio.CancelledError)
        else:
            assert call_results[idx] == idx


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "scheduler_cls",
    [ExitOrderedAsyncScheduler, KeySerializedAsyncScheduler],
)
async def test_messaging_server_error(scheduler_cls) -> None:
    done = asyncio.Event()
    total = 200
    error_idxs = [False] * 130 + [True] * 70
    random.shuffle(error_idxs)
    call_results: List[int | BaseException] = []
    return_results: List[int] = []
    # a list of events to make fucntions to return in the reversed order
    order_events = [asyncio.Event() for _ in range(total)]

    async def func(request: RPCMessage) -> int:
        # waits a short delay so that it return in the reversed order
        body = cast(Mapping[str, int], request.body)
        total = body["total"]
        idx = body["idx"]
        if idx == 0:
            await asyncio.sleep(0.01)
            for j in range(total, 0, -1):
                order_events[j - 1].set()
                await asyncio.sleep(0.01)  # let the next waiter proceed
        await order_events[idx].wait()
        if error_idxs[idx]:
            raise ZeroDivisionError("oops")
        return idx

    async def requester(client) -> None:
        async def _do_request(idx: int) -> int:
            try:
                ret = await client.invoke(
                    "func",
                    {
                        "total": total,
                        "idx": idx,
                    },
                    order_key="mykey",
                )
            except asyncio.CancelledError:
                return_results.append(-1)
                raise
            else:
                return_results.append(ret)
            return ret

        tasks = []
        for idx in range(total):
            tasks.append(asyncio.create_task(_do_request(idx)))
        call_results.extend(await asyncio.gather(*tasks, return_exceptions=True))

    scheduler = scheduler_cls()
    server_task = asyncio.create_task(dummy_server(scheduler, func, done))
    client_task = asyncio.create_task(dummy_client(requester, done))
    await asyncio.wait([server_task, client_task])

    if isinstance(scheduler, KeySerializedAsyncScheduler):
        # check memory leak
        assert len(scheduler._pending) == 0
        assert len(scheduler._tasks) == 0
        assert len(scheduler._futures) == 0
    elif isinstance(scheduler, ExitOrderedAsyncScheduler):
        assert len(scheduler._tasks) == 0

    for idx in range(total):
        if error_idxs[idx]:
            assert isinstance(call_results[idx], RPCUserError)
            e = cast(RPCUserError, call_results[idx])
            assert e.args[0] == "ZeroDivisionError"
        else:
            assert call_results[idx] == idx



"""
During the testing, you can launch multiple publishers
simultaneously, so as to check whether messages from
multiple publishers are distributed among the consumers.
"""

import asyncio
import json
import logging
import os
import random
import secrets

from callosum.lower.dispatch_redis import DispatchRedisTransport, RedisStreamAddress
from callosum.pubsub import Publisher


async def publish() -> None:
    redis_host = os.environ.get("REDIS_HOST", "127.0.0.1")
    redis_port = int(os.environ.get("REDIS_PORT", "6379"))
    pub = Publisher(
        bind=RedisStreamAddress(f"redis://{redis_host}:{redis_port}", "events"),
        serializer=lambda d: json.dumps(d).encode("utf8"),
        transport=DispatchRedisTransport,
    )
    agent_id = secrets.token_hex(2)  # publisher id

    async def heartbeats():
        for _ in range(3):
            await asyncio.sleep(1)
            msg_body = {
                "type": "instance_heartbeat",
                "agent_id": agent_id,
            }
            pub.push(msg_body)
            print("pushed heartbeat")

    async def addition_event(addend1: int, addend2: int):
        await asyncio.sleep(1.5)
        msg_body = {
            "type": "number_addition",
            "addends": (addend1, addend2),
        }
        pub.push(msg_body)
        print("pushed addition event")

    async with pub:
        task1 = asyncio.create_task(heartbeats())
        addend1 = random.randint(1, 10)
        addend2 = random.randint(10, 20)
        task2 = asyncio.create_task(addition_event(addend1, addend2))
        await asyncio.gather(task1, task2)
    print("publisher done")


if __name__ == "__main__":
    logging.basicConfig(
        format="%(asctime)s [%(levelname)s] %(name)s: %(message)s",
        level=logging.INFO,
    )
    log = logging.getLogger()
    asyncio.run(publish())



import asyncio
import json
import logging
import signal
import tracemalloc
from typing import Mapping, Type

import click

from callosum.lower.zeromq import ZeroMQAddress, ZeroMQRPCTransport
from callosum.ordering import (
    AbstractAsyncScheduler,
    ExitOrderedAsyncScheduler,
    KeySerializedAsyncScheduler,
)
from callosum.rpc import Peer

scheduler_types: Mapping[str, Type[AbstractAsyncScheduler]] = {
    "key-serialized": KeySerializedAsyncScheduler,
    "exit-ordered": ExitOrderedAsyncScheduler,
}

last_snapshot = None
show_output = True
scheduler = None

log = logging.getLogger()


async def handle_echo(request):
    if show_output:
        print("handle_echo()")
    return {
        "received": request.body["sent"],
    }


async def handle_add(request):
    if show_output:
        print("handle_add()")
    return {
        "result": request.body["a"] + request.body["b"],
    }


async def handle_output(request):
    global show_output
    show_output = request.body["enabled"]


async def handle_show_memory_stat(request):
    global last_snapshot, scheduler
    last_snapshot = last_snapshot.filter_traces((
        tracemalloc.Filter(False, "<frozen importlib._bootstrap>"),
        tracemalloc.Filter(False, tracemalloc.__file__),
    ))
    new_snapshot = tracemalloc.take_snapshot().filter_traces((
        tracemalloc.Filter(False, "<frozen importlib._bootstrap>"),
        tracemalloc.Filter(False, tracemalloc.__file__),
    ))
    top_stats = new_snapshot.compare_to(last_snapshot, "lineno")
    last_snapshot = new_snapshot
    print("[ Top 10 differences ]")
    for stat in top_stats[:10]:
        print(stat)
    print("[ Scheduler Queue Status ]")
    if isinstance(scheduler, KeySerializedAsyncScheduler):
        print("_tasks", len(scheduler._tasks))
        print("_futures", len(scheduler._futures))
        if hasattr(scheduler, "_pending"):
            print("_pending", len(scheduler._pending))
            print(scheduler._pending)
    elif isinstance(scheduler, ExitOrderedAsyncScheduler):
        print("_tasks", len(scheduler._tasks))


async def handle_long_delay(request):
    if show_output:
        print("handle_long_delay()")
    try:
        await asyncio.sleep(5)
        return {
            "received": request.body["sent"],
        }
    except asyncio.CancelledError:
        if show_output:
            print(" -> cancelled as expected")
        # NOTE: due to strange behaviour of asyncio, I have to reraise
        # otherwise, the task.cancelled() returns False
        raise


async def handle_error(request):
    if show_output:
        print("handle_error()")
    await asyncio.sleep(0.1)
    raise ZeroDivisionError("ooops")


async def serve(scheduler_type: str) -> None:
    global last_snapshot, scheduler

    sched_cls = scheduler_types[scheduler_type]
    scheduler = sched_cls()
    peer = Peer(
        bind=ZeroMQAddress("tcp://127.0.0.1:5020"),
        transport=ZeroMQRPCTransport,
        transport_opts={"attach_monitor": True},
        scheduler=scheduler,
        serializer=lambda o: json.dumps(o).encode("utf8"),
        deserializer=lambda b: json.loads(b),
    )
    peer.handle_function("echo", handle_echo)
    peer.handle_function("add", handle_add)
    peer.handle_function("long_delay", handle_long_delay)
    peer.handle_function("error", handle_error)
    peer.handle_function("set_output", handle_output)
    peer.handle_function("memstat", handle_show_memory_stat)

    loop = asyncio.get_running_loop()
    forever = loop.create_future()
    loop.add_signal_handler(signal.SIGINT, forever.cancel)
    loop.add_signal_handler(signal.SIGTERM, forever.cancel)

    try:
        tracemalloc.start(10)
        async with peer:
            last_snapshot = tracemalloc.take_snapshot()
            try:
                log.info("server started")
                await forever
            except asyncio.CancelledError:
                pass
        log.info("server terminated")
    finally:
        tracemalloc.stop()


@click.command()
@click.argument(
    "scheduler_type",
    default="exit-ordered",
    type=click.Choice([*scheduler_types.keys()]),
)
def main(scheduler_type):
    asyncio.run(serve(scheduler_type))


if __name__ == "__main__":
    logging.basicConfig(
        format="%(asctime)s [%(levelname)s] %(name)s: %(message)s",
        level=logging.INFO,
    )
    main()



import asyncio
import json
import logging
import os
import random
import secrets
import sys
import textwrap
import traceback

from callosum.lower.zeromq import ZeroMQAddress, ZeroMQRPCTransport
from callosum.rpc import Peer, RPCUserError

log = logging.getLogger()


async def test_simple(peer, initial_delay: float = 0):
    await asyncio.sleep(initial_delay)

    sent_token = secrets.token_hex(16)
    response = await peer.invoke(
        "echo",
        {
            "sent": sent_token,
        },
    )
    print(f"echoed {response['received']}")
    assert response["received"] == sent_token

    a = random.randint(1, 10)
    b = random.randint(10, 20)
    response = await peer.invoke(
        "add",
        {
            "a": a,
            "b": b,
        },
    )
    print(f"{a} + {b} = {response['result']}")
    assert response["result"] == a + b


async def test_timeout(peer):
    try:
        async with asyncio.timeout(0.5):
            await peer.invoke(
                "long_delay",
                {
                    "sent": "some-text",
                },
            )
    except asyncio.TimeoutError:
        print(
            "long_delay(): timeout occurred as expected " "(with per-call timeout)"
        )
    except Exception as e:
        print("long_delay(): unexpected error", e)
        sys.exit(1)
    else:
        print("long_delay(): timeout did not occur!")
        sys.exit(1)

    try:
        await peer.invoke(
            "long_delay",
            {
                "sent": "some-text",
            },
        )
    except asyncio.TimeoutError:
        print("long_delay(): timeout occurred as expected " "(with default timeout)")
    except Exception as e:
        print("long_delay(): unexpected error", e)
        sys.exit(1)
    else:
        print("long_delay(): timeout did not occur!")
        sys.exit(1)


async def test_exception(peer):
    try:
        await peer.invoke("error", {})
    except RPCUserError as e:
        print("error(): catched remote error as expected")
        print(textwrap.indent(e.traceback, prefix="| "))
    except Exception:
        print("error(): did not raise RPCUserError but strange one")
        traceback.print_exc()
        sys.exit(1)
    else:
        print("error(): did not return exception!")
        sys.exit(1)


async def single_client() -> None:
    peer = Peer(
        connect=ZeroMQAddress("tcp://localhost:5020"),
        transport=ZeroMQRPCTransport,
        transport_opts={"attach_monitor": True},
        serializer=lambda o: json.dumps(o).encode("utf8"),
        deserializer=lambda b: json.loads(b),
        invoke_timeout=2.0,
    )
    async with peer:
        await test_simple(peer)
        await test_exception(peer)
        await test_timeout(peer)
        await test_simple(peer)


async def overlapped_requests() -> None:
    peer = Peer(
        connect=ZeroMQAddress("tcp://localhost:5020"),
        transport=ZeroMQRPCTransport,
        transport_opts={"attach_monitor": True},
        serializer=lambda o: json.dumps(o).encode("utf8"),
        deserializer=lambda b: json.loads(b),
        invoke_timeout=2.0,
    )
    async with peer:
        t1 = asyncio.create_task(test_simple(peer))
        t2 = asyncio.create_task(test_simple(peer))
        t3 = asyncio.create_task(test_timeout(peer))
        t4 = asyncio.create_task(test_exception(peer))
        t5 = asyncio.create_task(test_timeout(peer))
        t6 = asyncio.create_task(test_simple(peer, initial_delay=0.5))
        t7 = asyncio.create_task(test_simple(peer, initial_delay=1.0))
        results = await asyncio.gather(
            t1, t2, t3, t4, t5, t6, t7, return_exceptions=True
        )
        print(results)
        if any(map(lambda r: isinstance(r, Exception), results)):
            print("There were failed clients.")
            sys.exit(1)


async def multi_clients() -> None:
    peer = Peer(
        connect=ZeroMQAddress("tcp://localhost:5020"),
        transport=ZeroMQRPCTransport,
        transport_opts={"attach_monitor": True},
        serializer=lambda o: json.dumps(o).encode("utf8"),
        deserializer=lambda b: json.loads(b),
        invoke_timeout=2.0,
    )
    async with peer:
        await peer.invoke("memstat", {})
        await peer.invoke("set_output", {"enabled": False})

        with open(os.devnull, "w", encoding="utf8") as out:
            orig_stdout, sys.stdout = sys.stdout, out
            try:
                for _ in range(10):
                    c1 = asyncio.create_task(single_client())
                    c2 = asyncio.create_task(single_client())
                    c3 = asyncio.create_task(overlapped_requests())
                    await asyncio.gather(c1, c2, c3, return_exceptions=True)
            finally:
                sys.stdout = orig_stdout

        await peer.invoke("set_output", {"enabled": True})
        await peer.invoke("memstat", {})


if __name__ == "__main__":
    logging.basicConfig(
        format="%(asctime)s [%(levelname)s] %(name)s: %(message)s",
        level=logging.INFO,
    )

    print("==== Testing with a single client ====")
    asyncio.run(single_client())

    print("==== Testing with overlapped requests ====")
    asyncio.run(overlapped_requests())

    print("==== Testing with multiple clients ====")
    asyncio.run(multi_clients())



import asyncio
import json
import logging
import signal

from callosum.lower.zeromq import ZeroMQAddress, ZeroMQRPCTransport
from callosum.ordering import ExitOrderedAsyncScheduler
from callosum.rpc import Peer


async def handle_echo(request):
    print("echo start")
    await asyncio.sleep(1)
    print("echo done")
    return {
        "received": request.body["sent"],
    }


async def handle_add(request):
    print("add start")
    await asyncio.sleep(0.5)
    print("add done")
    return {
        "result": request.body["a"] + request.body["b"],
    }


async def handle_delimeter(request):
    print("------")


async def serve():
    peer = Peer(
        bind=ZeroMQAddress("tcp://127.0.0.1:5010"),
        transport=ZeroMQRPCTransport,
        scheduler=ExitOrderedAsyncScheduler(),
        serializer=json.dumps,
        deserializer=json.loads,
    )
    peer.handle_function("echo", handle_echo)
    peer.handle_function("add", handle_add)
    peer.handle_function("print_delim", handle_delimeter)

    print("echo() will take 1 second and add() will take 0.5 second.")
    print(
        "You can confirm the effect of scheduler "
        "and the ordering key by the console logs.\n"
    )

    loop = asyncio.get_running_loop()
    forever = loop.create_future()
    loop.add_signal_handler(signal.SIGINT, forever.cancel)
    loop.add_signal_handler(signal.SIGTERM, forever.cancel)
    async with peer:
        try:
            print("server started")
            await forever
        except asyncio.CancelledError:
            pass
    print("server terminated")


if __name__ == "__main__":
    logging.basicConfig(
        format="%(asctime)s [%(levelname)s] %(name)s: %(message)s",
        level=logging.INFO,
    )
    log = logging.getLogger()
    asyncio.run(serve())



from typing import Optional


class CallosumError(Exception):
    pass


class ConfigurationError(CallosumError):
    pass


class AuthenticationError(CallosumError):
    pass


class InvalidAddressError(CallosumError, ValueError):
    def __init__(self, invalid_param: Optional[str] = None):
        if invalid_param:
            self.message = f"""
                {invalid_param} either must not be specified
                or is invalid.
                """
        else:
            self.message = "An invalid address was specified."


class ClientError(CallosumError):
    pass


class ServerError(CallosumError):
    pass
